from __future__ import annotations

import logging
import time

from dbetto import AttrsDict

log = logging.getLogger(__name__)


class ProfileDict(AttrsDict):
    """A class to store the results of time profiling."""

    def update_field(self, name: str, time_start: float) -> None:
        """Update the stored time.

        Parameters
        ----------
        name
            the name of the field to update. If it contains / this
            will be interpreted as subdictionaries.
        time_start
            the starting time of the block to evaluate
        """
        name_split = name.split("/")
        group = None
        dict_tmp = None

        time_end = time.time()

        for idx, name_tmp in enumerate(name_split):
            dict_tmp = self if (group is None) else dict_tmp[group]

            # if we are at the end and the name is not in the dictionary add it
            if (idx == len(name_split) - 1) and (name_tmp not in dict_tmp):
                dict_tmp[name_tmp] = time_end - time_start

            # append the time different
            elif (idx == len(name_split) - 1) and (name_tmp in dict_tmp):
                dict_tmp[name_tmp] = dict_tmp[name_tmp] + (time_end - time_start)

            # create a subdictionary
            elif name_tmp not in dict_tmp:
                dict_tmp[name_tmp] = {}

            group = name_tmp

    def __repr__(self):
        return f"ProfileDict({dict(self)})"

    def __str__(self):
        """Return a human-readable profiling summary."""
        return "\nReboost post processing took: \n" + self._format(self, indent=1)

    def _format(self, data: ProfileDict, indent: int = 1) -> str:
        """Recursively format the dictionary.

        Parameters
        ----------
        data
            The dictionary to format.
        indent
            The current indentation level.

        Returns
        -------
        the formatted print out.
        """
        output = ""
        space = " " * indent  # Indentation spaces

        for key, value in data.items():
            if isinstance(value, dict):  # If the value is a dictionary, recurse
                output += f"{space}- {key}:\n" + self._format(value, indent + 2)
            else:
                # Round floats to 1 decimal place
                value_print = round(value, 1) if isinstance(value, float) else value
                value_print = f"{value_print}".rjust(7) if value_print > 0 else "< 0.1".rjust(7)
                output += f"{space}- {key}".ljust(25)
                output += f": {value_print} s\n"

        return output



from __future__ import annotations

import logging
from typing import Any

import pint
import pyg4ometry as pg4
from lgdo import LGDO

log = logging.getLogger(__name__)

ureg = pint.get_application_registry()
"""The physical units registry."""

# default pretty printing of physical units
ureg.formatter.default_format = "~P"


def pg4_to_pint(obj) -> pint.Quantity:
    """Convert pyg4ometry object to pint Quantity."""
    if isinstance(obj, pint.Quantity):
        return obj
    if isinstance(obj, pg4.gdml.Defines.VectorBase):
        return [getattr(obj, field).eval() for field in ("x", "y", "z")] * ureg(obj.unit)
    msg = f"I don't know how to convert object of type {type(obj)} to pint object"
    raise ValueError(msg)


def units_convfact(data: Any, target_units: pint.Units) -> float:
    """Calculate numeric conversion factor to reach `target_units`.

    Parameters
    ----------
    data
        starting data structure. If an LGDO, try to determine units by peeking
        into its attributes. Otherwise, just return 1.
    target_units
        units you wish to convert data to.
    """
    if isinstance(data, LGDO) and "units" in data.attrs:
        return ureg(data.attrs["units"]).to(target_units).magnitude
    return 1


def unwrap_lgdo(data: Any, library: str = "ak") -> tuple(Any, pint.Unit | None):
    """Return a view of the data held by the LGDO and its physical units.

    Parameters
    ----------
    data
        the data container. If not an LGDO, it will be returned as is with
        ``None`` units.
    library
        forwarded to :func:`lgdo.view_as`.

    Returns
    -------
    A tuple of the un-lgdo'd data and the data units.
    """
    ret_data = data
    ret_units = None
    if isinstance(data, LGDO):
        ret_data = data.view_as(library)
        if "units" in data.attrs:
            ret_units = ureg(data.attrs["units"]).u

    return ret_data, ret_units


def unit_to_lh5_attr(unit: pint.Unit) -> str:
    """Convert Pint unit to a string that can be used as attrs["units"] in an LGDO."""
    # TODO: we should check if this can be always parsed by Unitful.jl
    if isinstance(unit, pint.Unit):
        return f"{unit:~C}"
    return unit



from __future__ import annotations

import awkward as ak
import numpy as np
from lgdo import Table

import reboost
from reboost.shape import cluster, group


def test_evtid_group():
    in_arr_evtid = ak.Array(
        {"evtid": [1, 1, 1, 2, 2, 10, 10, 11, 12, 12, 12], "time": np.zeros(11)}
    )

    # test with table input

    in_tab = Table(in_arr_evtid)
    out = group.group_by_evtid(in_tab)
    out_ak = out.view_as("ak")
    assert ak.all(out_ak.evtid == [[1, 1, 1], [2, 2], [10, 10], [11], [12, 12, 12]])
    assert ak.all(out_ak.time == [[0, 0, 0], [0, 0], [0, 0], [0], [0, 0, 0]])

    # test with ak.Array input
    out = group.group_by_evtid(in_tab.view_as("ak"))
    out_ak = out.view_as("ak")
    assert ak.all(out_ak.evtid == [[1, 1, 1], [2, 2], [10, 10], [11], [12, 12, 12]])
    assert ak.all(out_ak.time == [[0, 0, 0], [0, 0], [0, 0], [0], [0, 0, 0]])

    # test the eval in build hit also
    out_eval = reboost.core.evaluate_hit_table_layout(
        in_tab,
        "reboost.shape.group.group_by_evtid(STEPS)",
    )

    out_eval_ak = out_eval.view_as("ak")

    assert ak.all(out_ak.evtid == out_eval_ak.evtid)
    assert ak.all(out_ak.time == out_eval_ak.time)


def test_time_group():
    # time units are ns
    in_arr_evtid = ak.Array(
        {
            "evtid": [1, 1, 1, 2, 2, 2, 2, 2, 11, 12, 12, 12, 15, 15, 15, 15, 15],
            "time": [
                0,
                -2000,
                3000,
                0,
                100,
                1200,
                17000,
                17010,
                0,
                0,
                0,
                -5000,
                150,
                151,
                152,
                3000,
                3100,
            ],
        }
    )

    in_tab = Table(in_arr_evtid)

    # 1us =1000ns
    out = group.group_by_time(in_tab, window=1)
    out_ak = out.view_as("ak")
    assert ak.all(
        out_ak.evtid
        == [[1], [1], [1], [2, 2], [2], [2, 2], [11], [12], [12, 12], [15, 15, 15], [15, 15]]
    )
    assert ak.all(
        out_ak.time
        == [
            [-2000],
            [0],
            [3000],
            [0, 100],
            [1200],
            [17000, 17010],
            [0],
            [-5000],
            [0, 0],
            [150, 151, 152],
            [3000, 3100],
        ]
    )


def test_cluster_basic():
    trackid = ak.Array([[1, 1, 1, 2, 2, 3, 3, 7], [2, 2, 2, 3, 3, 3], [1]])

    runs = ak.run_lengths(trackid)
    assert ak.all(
        cluster.apply_cluster(runs, trackid).view_as("ak")
        == ak.Array([[[1, 1, 1], [2, 2], [3, 3], [7]], [[2, 2, 2], [3, 3, 3]], [[1]]])
    )


def test_cluster_by_step_length():
    # test wrapped function
    trackid = ak.Array([[1, 1, 1, 2, 2, 2, 2, 2], [2, 2, 2, 3, 3, 3], [1]])
    x = ak.Array([[0, 0.1, 0.5, 1, 2, 2.01, 2.02, 4], [0, 1, 4, 5, 5, 6], [0]])
    y = ak.Array([[0, 0, 0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0], [0]])
    z = ak.Array([[0, 0, 0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0], [0]])
    dist = ak.Array([[0.1, 0.1, 0.1, 0.1, 2, 2, 2, 2], [2, 2, 2, 2, 2, 2], [0.3]])

    clusters = cluster.cluster_by_step_length(
        trackid, x, y, z, dist, threshold=0.2, threshold_surf=0.02, surf_cut=0.15
    )

    assert ak.all(clusters.view_as("ak") == ak.Array([[1, 1, 1, 1, 3, 1], [1, 1, 1, 2, 1], [1]]))

    # test without surface region

    clusters = cluster.cluster_by_step_length(trackid, x, y, z, threshold=0.2)
    assert ak.all(clusters.view_as("ak") == ak.Array([[2, 1, 1, 3, 1], [1, 1, 1, 2, 1], [1]]))


def test_step_length():
    trackid = ak.Array([[1, 1, 1, 2, 2, 3, 3, 7], [2, 2, 2, 3, 3, 3], [1]])
    x = ak.Array([[0, 0, 1, 1, 2, 2, 3, 4], [0, 1, 4, 5, 5, 6], [0]])
    y = ak.Array([[0, 0, 0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0], [0]])
    z = ak.Array([[0, 0, 0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0], [0]])

    run = ak.run_lengths(trackid)

    x_cluster = cluster.apply_cluster(run, x).view_as("ak")
    y_cluster = cluster.apply_cluster(run, y).view_as("ak")
    z_cluster = cluster.apply_cluster(run, z).view_as("ak")

    steps = cluster.step_lengths(x_cluster, y_cluster, z_cluster).view_as("ak")

    for idx, sub_array in enumerate(steps):
        assert ak.all(sub_array == ak.Array([[[0, 1], [1], [1], []], [[1, 3], [0, 1.0]], []])[idx])



from __future__ import annotations

import logging

import legendoptics.scintillate as sc
import numba
import numpy as np
import pint
from legendoptics import lar
from lgdo import lh5
from lgdo.lh5 import LH5Iterator
from lgdo.types import Array, Histogram, Table
from numba import njit, prange
from numpy.lib.recfunctions import structured_to_unstructured

from .numba_pdg import numba_pdgid_funcs

log = logging.getLogger(__name__)


OPTMAP_ANY_CH = -1
OPTMAP_SUM_CH = -2


def open_optmap(optmap_fn: str):
    maps = lh5.ls(optmap_fn)
    # TODO: rewrite logic to only accept _<number> instead of a blacklist
    det_ntuples = [m for m in maps if m not in ("all", "_hitcounts", "_hitcounts_exp", "all_orig")]
    detids = np.array([int(m.lstrip("_")) for m in det_ntuples])
    detidx = np.arange(0, detids.shape[0])

    optmap_all = lh5.read("/all/p_det", optmap_fn)
    assert isinstance(optmap_all, Histogram)
    optmap_edges = tuple([b.edges for b in optmap_all.binning])

    ow = np.empty((detidx.shape[0] + 2, *optmap_all.weights.nda.shape), dtype=np.float64)
    # 0, ..., len(detidx)-1 AND OPTMAP_ANY_CH might be negative.
    ow[OPTMAP_ANY_CH] = optmap_all.weights.nda
    for i, nt in zip(detidx, det_ntuples):
        optmap = lh5.read(f"/{nt}/p_det", optmap_fn)
        assert isinstance(optmap, Histogram)
        ow[i] = optmap.weights.nda

    # if we have any individual channels registered, the sum is potentially larger than the
    # probability to find _any_ hit.
    if len(detidx) != 0:
        ow[OPTMAP_SUM_CH] = np.sum(ow[0:-2], axis=0, where=(ow[0:-2] >= 0))
        assert not np.any(ow[OPTMAP_SUM_CH] < 0)
    else:
        detidx = np.array([OPTMAP_ANY_CH])
        detids = np.array([0])
        ow[OPTMAP_SUM_CH] = ow[OPTMAP_ANY_CH]

    # give this check some numerical slack.
    if np.any(
        ow[OPTMAP_SUM_CH][ow[OPTMAP_ANY_CH] >= 0] - ow[OPTMAP_ANY_CH][ow[OPTMAP_ANY_CH] >= 0]
        < -1e-15
    ):
        msg = "optical map does not fulfill relation sum(p_i) >= p_any"
        raise ValueError(msg)

    # get the exponent from the optical map file
    optmap_multi_det_exp = lh5.read("/_hitcounts_exp", optmap_fn).value
    assert isinstance(optmap_multi_det_exp, float)

    return detids, detidx, optmap_edges, ow, optmap_multi_det_exp


def iterate_stepwise_depositions(
    edep_df: np.rec.recarray,
    optmap_for_convolve,
    scint_mat_params: sc.ComputedScintParams,
    rng: np.random.Generator = None,
    dist: str = "multinomial",
    mode: str = "no-fano",
):
    # those np functions are not supported by numba, but needed for efficient array access below.
    x0 = structured_to_unstructured(edep_df[["xloc_pre", "yloc_pre", "zloc_pre"]], np.float64)
    x1 = structured_to_unstructured(edep_df[["xloc_post", "yloc_post", "zloc_post"]], np.float64)

    rng = np.random.default_rng() if rng is None else rng
    output_map, res = _iterate_stepwise_depositions(
        edep_df, x0, x1, rng, *optmap_for_convolve, scint_mat_params, dist, mode
    )
    if res["any_no_stats"] > 0 or res["det_no_stats"] > 0:
        log.warning(
            "had edep out in voxels without stats: %d (%.2f%%)",
            res["any_no_stats"],
            res["det_no_stats"],
        )
    if res["oob"] > 0:
        log.warning(
            "had edep out of map bounds: %d (%.2f%%)",
            res["oob"],
            (res["oob"] / (res["ib"] + res["oob"])) * 100,
        )
    log.debug(
        "VUV_primary %d ->hits_any %d ->hits %d (%.2f %% primaries detected)",
        res["vuv_primary"],
        res["hits_any"],
        res["hits"],
        (res["hits_any"] / res["vuv_primary"]) * 100,
    )
    log.debug("hits/hits_any %.2f", res["hits"] / res["hits_any"])
    return output_map


_pdg_func = numba_pdgid_funcs()


@njit
def _pdgid_to_particle(pdgid: int) -> sc.ParticleIndex:
    abs_pdgid = abs(pdgid)
    if abs_pdgid == 1000020040:
        return sc.PARTICLE_INDEX_ALPHA
    if abs_pdgid == 1000010020:
        return sc.PARTICLE_INDEX_DEUTERON
    if abs_pdgid == 1000010030:
        return sc.PARTICLE_INDEX_TRITON
    if _pdg_func.is_nucleus(pdgid):
        return sc.PARTICLE_INDEX_ION
    return sc.PARTICLE_INDEX_ELECTRON


__counts_per_bin_key_type = numba.types.UniTuple(numba.types.int64, 3)


# - run with NUMBA_FULL_TRACEBACKS=1 NUMBA_BOUNDSCHECK=1 for testing/checking
# - cache=True does not work with outer prange, i.e. loading the cached file fails (numba bug?)
# - the output dictionary is not threadsafe, so parallel=True is not working with it.
@njit(parallel=False, nogil=True, cache=True)
def _iterate_stepwise_depositions(
    edep_df,
    x0,
    x1,
    rng,
    detids,
    detidx,
    optmap_edges,
    optmap_weights,
    optmap_multi_det_exp,
    scint_mat_params: sc.ComputedScintParams,
    dist: str,
    mode: str,
):
    pdgid_map = {}
    output_map = {}
    oob = ib = ph_cnt = ph_det = ph_det2 = any_no_stats = det_no_stats = 0  # for statistics
    for rowid in prange(edep_df.shape[0]):
        # if rowid % 100000 == 0:
        #     print(rowid)
        t = edep_df[rowid]

        # get the particle information.
        if t.particle not in pdgid_map:
            pdgid_map[t.particle] = (_pdgid_to_particle(t.particle), _pdg_func.charge(t.particle))

        # do the scintillation.
        part, charge = pdgid_map[t.particle]
        scint_times = sc.scintillate(
            scint_mat_params,
            x0[rowid],
            x1[rowid],
            t.v_pre,
            t.v_post,
            t.time,
            part,
            charge,
            t.edep,
            rng,
            emission_term_model=("poisson" if mode == "no-fano" else "normal_fano"),
        )
        if scint_times.shape[0] == 0:  # short-circuit if we have no photons at all.
            continue
        ph_cnt += scint_times.shape[0]

        # coordinates -> bins of the optical map.
        bins = np.empty((scint_times.shape[0], 3), dtype=np.int64)
        for j in range(3):
            bins[:, j] = np.digitize(scint_times[:, j + 1], optmap_edges[j])
            # normalize all out-of-bounds bins just to one end.
            bins[:, j][bins[:, j] == optmap_edges[j].shape[0]] = 0

        # there are _much_ less unique bins, unfortunately np.unique(..., axis=n) does not work
        # with numba; also np.sort(..., axis=n) also does not work.

        counts_per_bin = numba.typed.Dict.empty(
            key_type=__counts_per_bin_key_type,
            value_type=np.int64,
        )

        # get probabilities from map.
        hitcount = np.zeros((detidx.shape[0], bins.shape[0]), dtype=np.int64)
        for j in prange(bins.shape[0]):
            # note: subtract 1 from bins, to account for np.digitize output.
            cur_bins = (bins[j, 0] - 1, bins[j, 1] - 1, bins[j, 2] - 1)
            if cur_bins[0] == -1 or cur_bins[1] == -1 or cur_bins[2] == -1:
                oob += 1
                continue  # out-of-bounds of optmap
            ib += 1

            px_any = optmap_weights[OPTMAP_ANY_CH, cur_bins[0], cur_bins[1], cur_bins[2]]
            if px_any < 0.0:
                any_no_stats += 1
                continue
            if px_any == 0.0:
                continue

            if dist == "multinomial":
                if rng.uniform() >= px_any:
                    continue
                ph_det += 1
                # we detect this energy deposition; we should at least get one photon out here!

                detsel_size = 1
                if np.isfinite(optmap_multi_det_exp):
                    detsel_size = rng.geometric(1 - np.exp(-optmap_multi_det_exp))

                px_sum = optmap_weights[OPTMAP_SUM_CH, cur_bins[0], cur_bins[1], cur_bins[2]]
                assert px_sum >= 0.0  # should not be negative.
                detp = np.empty(detidx.shape, dtype=np.float64)
                had_det_no_stats = 0
                for d in detidx:
                    # normalize so that sum(detp) = 1
                    detp[d] = optmap_weights[d, cur_bins[0], cur_bins[1], cur_bins[2]] / px_sum
                    if detp[d] < 0.0:
                        had_det_no_stats = 1
                        detp[d] = 0.0
                det_no_stats += had_det_no_stats

                # should be equivalent to rng.choice(detidx, size=(detsel_size, p=detp)
                detsel = detidx[
                    np.searchsorted(np.cumsum(detp), rng.random(size=(detsel_size,)), side="right")
                ]
                for d in detsel:
                    hitcount[d, j] += 1
                ph_det2 += detsel.shape[0]

            elif dist == "poisson":
                # store the photon count in each bin, to sample them all at once below.
                if cur_bins not in counts_per_bin:
                    counts_per_bin[cur_bins] = 1
                else:
                    counts_per_bin[cur_bins] += 1

            else:
                msg = "unknown distribution"
                raise RuntimeError(msg)

        if dist == "poisson":
            for j, (cur_bins, ph_counts_to_poisson) in enumerate(counts_per_bin.items()):
                had_det_no_stats = 0
                had_any = 0
                for d in detidx:
                    detp = optmap_weights[d, cur_bins[0], cur_bins[1], cur_bins[2]]
                    if detp < 0.0:
                        had_det_no_stats = 1
                        continue
                    pois_cnt = rng.poisson(lam=ph_counts_to_poisson * detp)
                    hitcount[d, j] += pois_cnt
                    ph_det2 += pois_cnt
                    had_any = 1
                ph_det += had_any
                det_no_stats += had_det_no_stats

        assert scint_times.shape[0] >= hitcount.shape[1]  # TODO: use the right assertion here.
        out_hits_len = np.sum(hitcount)
        if out_hits_len > 0:
            out_times = np.empty(out_hits_len, dtype=np.float64)
            out_det = np.empty(out_hits_len, dtype=np.int64)
            out_idx = 0
            for d in detidx:
                hc_d_plane_max = np.max(hitcount[d, :])
                # untangle the hitcount array in "planes" that only contain the given number of hits per
                # channel. example: assume a "histogram" of hits per channel:
                #     x |   |    <-- this is plane 2 with 1 hit ("max plane")
                #     x |   | x  <-- this is plane 1 with 2 hits
                # ch: 1 | 2 | 3
                for hc_d_plane_cnt in range(1, hc_d_plane_max + 1):
                    hc_d_plane = hitcount[d, :] >= hc_d_plane_cnt
                    hc_d_plane_len = np.sum(hc_d_plane)
                    if hc_d_plane_len == 0:
                        continue

                    # note: we assume "immediate" propagation after scintillation. Here, a single timestamp
                    # might be coipied to output/"detected" twice.
                    out_times[out_idx : out_idx + hc_d_plane_len] = scint_times[hc_d_plane, 0]
                    out_det[out_idx : out_idx + hc_d_plane_len] = detids[d]
                    out_idx += hc_d_plane_len
            assert out_idx == out_hits_len  # ensure that all of out_{det,times} is filled.
            output_map[np.int64(rowid)] = (t.evtid, out_det, out_times)

    stats = {
        "oob": oob,
        "ib": ib,
        "vuv_primary": ph_cnt,
        "hits_any": ph_det,
        "hits": ph_det2,
        "any_no_stats": any_no_stats,
        "det_no_stats": det_no_stats,
    }
    return output_map, stats


def get_output_table(output_map):
    ph_count_o = 0
    for _rawid, (_evtid, det, _times) in output_map.items():
        ph_count_o += det.shape[0]

    out_idx = 0
    out_evtid = np.empty(ph_count_o, dtype=np.int64)
    out_det = np.empty(ph_count_o, dtype=np.int64)
    out_times = np.empty(ph_count_o, dtype=np.float64)
    for _rawid, (evtid, det, times) in output_map.items():
        o_len = det.shape[0]
        out_evtid[out_idx : out_idx + o_len] = evtid
        out_det[out_idx : out_idx + o_len] = det
        out_times[out_idx : out_idx + o_len] = times
        out_idx += o_len

    tbl = Table({"evtid": Array(out_evtid), "det_uid": Array(out_det), "time": Array(out_times)})
    return ph_count_o, tbl


def convolve(
    map_file: str,
    edep_file: str,
    edep_path: str,
    material: str,
    output_file: str | None = None,
    buffer_len: int = int(1e6),
    dist_mode: str = "multinomial+no-fano",
):
    if material not in ["lar", "pen"]:
        msg = f"unknown material {material} for scintillation"
        raise ValueError(msg)

    if material == "lar":
        scint_mat_params = sc.precompute_scintillation_params(
            lar.lar_scintillation_params(),
            lar.lar_lifetimes().as_tuple(),
        )
    elif material == "pen":
        scint_mat_params = sc.precompute_scintillation_params(
            lar.pen_scintillation_params(),
            (1 * pint.get_application_registry().ns),  # dummy!
        )

    log.info("opening map %s", map_file)
    optmap_for_convolve = open_optmap(map_file)

    # special handling of distributions and flags.
    dist, mode = dist_mode.split("+")
    assert dist in ("multinomial", "poisson")
    assert mode in ("", "no-fano")

    log.info("opening energy deposition hit output %s", edep_file)
    it = LH5Iterator(edep_file, edep_path, buffer_len=buffer_len)
    for it_count, (edep_lgdo, edep_events, edep_n_rows) in enumerate(it):
        assert (it_count == 0) == (edep_events == 0)
        edep_df = edep_lgdo.view_as("pd").iloc[0:edep_n_rows].to_records()

        log.info("start event processing (%d)", it_count)
        output_map = iterate_stepwise_depositions(
            edep_df, optmap_for_convolve, scint_mat_params, dist=dist, mode=mode
        )

        log.info("store output photon hits (%d)", it_count)
        ph_count_o, tbl = get_output_table(output_map)
        log.debug(
            "output photons: %d energy depositions -> %d photons", len(output_map), ph_count_o
        )
        if output_file is not None:
            lh5.write(tbl, "optical", lh5_file=output_file, group="stp", wo_mode="append")



from __future__ import annotations

from reboost import build_hit, core, iterator, math, shape
from reboost._version import version as __version__

__all__ = [
    "__version__",
    "build_glm",
    "build_hit",
    "build_hit",
    "build_tcm",
    "core",
    "iterator",
    "math",
    "optmap",
    "shape",
]



from __future__ import annotations

import logging

import awkward as ak
import numpy as np
from lgdo import Table, VectorOfVectors

log = logging.getLogger(__name__)


def _sort_data(obj: ak.Array, *, time_name: str = "time", evtid_name: str = "evtid") -> ak.Array:
    """Sort the data by evtid then time.

    Parameters
    ----------
    obj
        array of records containing fields `time` and `evtid`.
    time_name
        name of the time field in `obj`.
    evtid_name
        name of the evtid field in `obj`.

    Returns
    -------
    sorted awkward array
    """
    obj = obj[ak.argsort(obj[evtid_name])]
    obj_unflat = ak.unflatten(obj, ak.run_lengths(obj[evtid_name]))

    indices = ak.argsort(obj_unflat[time_name], axis=-1)
    sorted_obj = obj_unflat[indices]

    return ak.flatten(sorted_obj)


def group_by_evtid(data: Table | ak.Array, *, evtid_name: str = "evtid") -> Table:
    """Simple grouping by evtid.

    Takes the input `stp` :class:`lgdo.Table` from remage and defines groupings of steps (i.e the
    `cumulative_length` for a vector of vectors). This then defines the output table (also :class:`lgdo.Table`),
    on which processors can add fields.

    Parameters
    ----------
    data
        LGDO Table which must contain the `evtid` field.
    evtid_name
        the name of the index field in the input table.

    Returns
    -------
    LGDO table of :class:`VectorOfVector` for each field.

    Note
    ----
    The input table must be sorted (by `evtid`).
    """
    # convert to awkward
    obj_ak = data.view_as("ak") if isinstance(data, Table) else data

    # extract cumulative lengths
    counts = ak.run_lengths(obj_ak[evtid_name])
    cumulative_length = np.cumsum(counts)

    # convert to numpy
    if isinstance(cumulative_length, ak.Array):
        cumulative_length = cumulative_length.to_numpy()

    # build output table
    out_tbl = Table(size=len(cumulative_length))

    for f in obj_ak.fields:
        out_tbl.add_field(
            f,
            VectorOfVectors(
                cumulative_length=cumulative_length, flattened_data=obj_ak[f].to_numpy()
            ),
        )
    return out_tbl


def group_by_time(
    data: Table | ak.Array,
    window: float = 10,
    time_name: str = "time",
    evtid_name: str = "evtid",
    fields: list | None = None,
) -> Table:
    """Grouping of steps by `evtid` and `time`.

    Takes the input `stp` :class:`lgdo.Table` from remage and defines groupings of steps (i.e the
    `cumulative_length` for a vector of vectors). This then defines the output table (also :class:`lgdo.Table`),
    on which processors can add fields.

    The windowing is based on defining a new group when the `evtid` changes or when the time increases by `> window`,
    which is in units of us.

    Parameters
    ----------
    data
        :class:`lgdo.Table` or `ak.Array` which must contain the time_name and evtid_name fields
    window
        time window in us used to search for coincident hits
    time_name
        name of the timing field
    evtid_name
        name of the evtid field
    fields
        names of fields to include in the output table, if None includes all

    Returns
    -------
    LGDO table of :class:`VectorOfVector` for each field.

    Note
    ----
    The input table must be sorted (first by `evtid` then `time`).
    """
    obj = data.view_as("ak") if isinstance(data, Table) else data
    obj = _sort_data(obj, time_name=time_name, evtid_name=evtid_name)

    # get difference
    time_diffs = np.diff(obj[time_name])
    index_diffs = np.array(np.diff(obj[evtid_name]), dtype=np.int32)

    # index of the last element in each run
    time_change = (time_diffs > window * 1000) & (index_diffs == 0)
    index_change = index_diffs > 0

    # cumulative length is just the index of changes plus 1
    cumulative_length = np.array(np.where(time_change | index_change))[0] + 1

    # add the las grouping
    cumulative_length = np.append(cumulative_length, len(obj[time_name]))

    # convert to numpy
    if isinstance(cumulative_length, ak.Array):
        cumulative_length = cumulative_length.to_numpy()

    # build output table
    out_tbl = Table(size=len(cumulative_length))

    fields = obj.fields if fields is None else fields
    for f in fields:
        out_tbl.add_field(
            f,
            VectorOfVectors(cumulative_length=cumulative_length, flattened_data=obj[f].to_numpy()),
        )

    return out_tbl



from __future__ import annotations

import copy
import logging

import awkward as ak
import numpy as np
from lgdo import Array, Table, lh5
from lgdo.lh5 import LH5Iterator, LH5Store
from numpy.typing import ArrayLike

from reboost import utils

log = logging.getLogger(__name__)


def get_glm_rows(stp_evtids: ArrayLike, vert: ArrayLike, *, start_row: int = 0) -> ak.Array:
    """Get the rows of the Geant4 event lookup map (glm).

    Parameters
    ----------
    stp_evtids
        Array of evtids for the steps
    vert
        Array of simulated evtid for the vertices.
    start_row
        The index of the first element of stp_evtids.

    Returns
    -------
    an awkward array of the `glm`.
    """
    # convert inputs
    if stp_evtids is None:
        output = ak.Array({"evtid": vert})
        output["n_rows"] = np.array([0] * len(vert), dtype=float)
        output["start_row"] = np.array([np.nan] * len(vert), dtype=float)
        return output

    if not isinstance(stp_evtids, np.ndarray):
        stp_evtids = (
            stp_evtids.to_numpy() if isinstance(stp_evtids, ak.Array) else np.array(stp_evtids)
        )

    if not isinstance(vert, np.ndarray):
        vert = vert.to_numpy() if isinstance(vert, ak.Array) else np.array(vert)

    # check that the steps and vertices are sorted or the algorithm will fail

    if not np.all(vert[:-1] <= vert[1:]):
        msg = "The vertices must be sorted"
        raise ValueError(msg)

    if len(stp_evtids) > 0 and not np.all(stp_evtids[:-1] <= stp_evtids[1:]):
        msg = "The steps must be sorted"
        raise ValueError(msg)

    # convert to numpy
    stp_indices = np.arange(len(stp_evtids)) + start_row

    # cut the arrays
    sel = (stp_evtids >= vert[0]) & (stp_evtids <= vert[-1])
    stp_evtids = stp_evtids[sel]
    stp_indices = stp_indices[sel]

    # build output
    output = ak.Array({"evtid": vert})

    # restructure to jagged array
    counts = ak.run_lengths(stp_evtids)
    steps = ak.unflatten(stp_evtids, counts)
    indices = ak.unflatten(stp_indices, counts)

    ak_tmp = ak.Array({"evtid": ak.fill_none(ak.firsts(steps), np.nan), "indices": indices})

    # find indices to insert the new entries
    positions = np.searchsorted(output.evtid, ak_tmp.evtid)

    # add a check that every ak_tmp evtid is in output.evtid ?
    if not np.all(np.isin(ak_tmp.evtid, output.evtid)):
        bad_evtid = ak_tmp.evtid[~np.isin(ak_tmp.evtid, output.evtid)]
        msg = f"Error not every evtid in the stp table is present in the vertex table {bad_evtid} are not"
        raise ValueError(msg)

    # get the start row
    start_row = np.array([np.nan] * len(vert), dtype=float)
    start_row[positions] = ak.fill_none(ak.firsts(ak_tmp.indices), np.nan)

    n_row = np.array([0] * len(vert), dtype=float)
    n_row[positions] = ak.num(ak_tmp.indices)

    # add to the  output
    output["n_rows"] = n_row
    output["start_row"] = start_row

    return output


def get_stp_evtids(
    lh5_table: str,
    stp_file: str,
    id_name: str,
    start_row: int,
    last_vertex_evtid: int,
    stp_buffer: int,
) -> tuple[int, int, ak.Array]:
    """Extracts the rows of a stp file corresponding to a particular range of `evtid`.

    The reading starts at `start_row` to allow for iterating through
    the file. The iteration stops when the `evtid` being read is larger than
    `last_vertex_evtid`.

    Parameters
    ----------
    lh5_table
        the table name to read.
    stp_file
        the file name path.
    id_name
        the name of the `evtid` field.
    start_row
        the row to begin reading.
    last_vertex_evtid
        the last evtid to read up to.
    stp_buffer
        the number of rows to read at once.

    Returns
    -------
         a tuple of the updated `start_row`, the first row for the chunk and an awkward Array of the steps.
    """
    # make a LH5Store
    store = LH5Store()

    # some outputs
    evtids_proc = None
    last_evtid = 0
    chunk_start = 0

    # get the total number of rows
    n_rows_tot = store.read_n_rows(f"{lh5_table}/{id_name}", stp_file)

    # iterate over the file
    # stop when the entire file is read

    while start_row < n_rows_tot:
        # read the file
        lh5_obj, n_read = store.read(
            f"{lh5_table}/{id_name}",
            stp_file,
            start_row=start_row,
            n_rows=stp_buffer,
        )
        evtids = lh5_obj.view_as("ak")

        # pick the first evtid
        if evtids.ndim > 1:
            evtids = ak.fill_none(ak.firsts(evtids, axis=-1), -1)

        # if the evtids_proc is not set then this is the first valid chunk
        if evtids_proc is None:
            evtids_proc = evtids
            chunk_start = start_row
        elif evtids[0] <= last_vertex_evtid:
            evtids_proc = ak.concatenate((evtids_proc, evtids))

        # get the last evtid
        last_evtid = evtids[-1]

        # if the last evtid is greater than the last vertex we can stop reading
        if last_evtid > last_vertex_evtid or (start_row + n_read >= n_rows_tot):
            break

        # increase rhe start row for the next read

        if start_row + n_read <= n_rows_tot:
            start_row += n_read

    return start_row, chunk_start, evtids_proc


def build_glm(
    stp_files: str | list[str],
    glm_files: str | list[str] | None,
    *,
    out_table_name: str = "glm",
    id_name: str = "g4_evtid",
    evtid_buffer: int = int(1e7),
    stp_buffer: int = int(1e7),
) -> ak.Array | None:
    """Builds a g4_evtid look up (glm) from the stp data.

    This object is used by `reboost` to efficiency iterate through the data.
    It consists of a :class:`lgdo.VectorOfVectors` for each lh5_table in the input files.
    The rows of this :class:`lgdo.VectorOfVectors` correspond to the `id_name` while the data
    are the `stp` indices for this event.

    Parameters
    ----------
    stp_files
        path to the stp (input) file.
    glm_files
        path to the glm data, can also be `None` in which case an `ak.Array` is returned in memory.
    out_table_name
        name for the output table.
    id_name
        name of the evtid file, default `g4_evtid`.
    stp_buffer
        the number of rows of the step file to read at a time
    evtid_buffer
        the number of evtids to read at a time

    Returns
    -------
    either `None` or an `ak.Array`
    """
    store = LH5Store()
    files = utils.get_file_dict(stp_files=stp_files, glm_files=glm_files)

    # loop over files
    glm_sum = {}

    for file_idx, stp_file in enumerate(files.stp):
        msg = f"Start generating glm for {stp_file} "
        log.info(msg)

        # loop over the lh5_tables
        lh5_table_list = list(lh5.ls(stp_file, "stp/"))

        # get rows in the table
        if files.glm[file_idx] is None:
            for lh5_table in lh5_table_list:
                if lh5_table.replace("stp/", "") not in glm_sum:
                    glm_sum[lh5_table.replace("stp/", "")] = None
        else:
            glm_sum = None

        # start row for each table
        start_row = dict.fromkeys(lh5_table_list, 0)

        vfield = f"vtx/{id_name}"

        # iterate over the vertex table
        for vert_obj, vidx, n_evtid in LH5Iterator(stp_file, vfield, buffer_len=evtid_buffer):
            # range of vertices
            vert_ak = vert_obj.view_as("ak")[:n_evtid]

            msg = f"... read chunk {vidx}"
            log.debug(msg)

            for idx, lh5_table in enumerate(lh5_table_list):
                # create the output table
                out_tab = Table(size=len(vert_ak))

                # read the stp rows starting from `start_row` until the
                # evtid is larger than that in the vertices

                start_row_tmp, chunk_row, evtids = get_stp_evtids(
                    lh5_table,
                    stp_file,
                    id_name,
                    start_row[lh5_table],
                    last_vertex_evtid=vert_ak[-1],
                    stp_buffer=stp_buffer,
                )

                # set the start row for the next chunk
                start_row[lh5_table] = start_row_tmp

                # now get the glm rows
                glm = get_glm_rows(evtids, vert_ak, start_row=chunk_row)

                for field in ["evtid", "n_rows", "start_row"]:
                    out_tab.add_field(field, Array(glm[field].to_numpy()))

                # write the output file
                mode = "of" if (vidx == 0 and idx == 0) else "append"

                lh5_subgroup = lh5_table.replace("stp/", "")

                if files.glm[file_idx] is not None:
                    store.write(
                        out_tab,
                        f"{out_table_name}/{lh5_subgroup}",
                        files.glm[file_idx],
                        wo_mode=mode,
                    )
                else:
                    glm_sum[lh5_subgroup] = (
                        copy.deepcopy(glm)
                        if glm_sum[lh5_subgroup] is None
                        else ak.concatenate((glm_sum[lh5_subgroup], glm))
                    )
        msg = f"Finished generating glm for {stp_file} "
        log.info(msg)

    # return if it was requested to keep glm in memory
    if glm_sum is not None:
        return ak.Array(glm_sum)
    return None



from __future__ import annotations

import logging

import colorlog


def setup_log(level: int | None = None, multiproc: bool = False) -> None:
    """Setup a colored logger for this package.

    Parameters
    ----------
    level
        initial log level, or ``None`` to use the default.
    multiproc
        set to ``True`` to include process ID in log output (i.e. for multiprocessing setups)
    """
    fmt = "%(log_color)s%(name)s [%(levelname)s]"
    if multiproc:
        fmt += " (pid=%(process)s)"
    fmt += " %(message)s"

    handler = colorlog.StreamHandler()
    handler.setFormatter(colorlog.ColoredFormatter(fmt))

    logger = logging.getLogger("reboost")
    logger.addHandler(handler)
    if level is not None:
        logger.setLevel(level)



from __future__ import annotations

import logging

import awkward as ak
import numba
import numpy as np
from lgdo import VectorOfVectors

log = logging.getLogger(__name__)


def apply_cluster(
    cluster_run_lengths: VectorOfVectors | ak.Array, field: ak.Array | VectorOfVectors
) -> VectorOfVectors:
    """Apply clustering to a field.

    Parameters
    ----------
    cluster_ids
        run lengths of each cluster
    field
        the field to cluster
    """
    if isinstance(cluster_run_lengths, VectorOfVectors):
        cluster_run_lengths = cluster_run_lengths.view_as("ak")

    if isinstance(field, VectorOfVectors):
        field = field.view_as("ak")

    n_cluster = ak.num(cluster_run_lengths, axis=-1)
    clusters = ak.unflatten(ak.flatten(field), ak.flatten(cluster_run_lengths))

    # reshape into cluster oriented
    return VectorOfVectors(ak.unflatten(clusters, n_cluster))


def cluster_by_step_length(
    trackid: ak.Array | VectorOfVectors,
    pos_x: ak.Array | VectorOfVectors,
    pos_y: ak.Array | VectorOfVectors,
    pos_z: ak.Array | VectorOfVectors,
    dist: ak.Array | VectorOfVectors | None = None,
    surf_cut: float | None = None,
    threshold: float = 0.1,
    threshold_surf: float | None = None,
) -> VectorOfVectors:
    """Perform clustering based on the step length.

    Steps are clustered based on distance, if either:
     - a step is in a new track,
     - a step moves from surface to bulk region (or visa versa),
     - the distance between the current step and the first step of the current cluster is above a threshold.

    Then a new cluster is started. The surface region is defined as the volume
    less than surf_cut distance to the surface. This allows for a fine tuning of the
    parameters to be different for bulk and surface.

    Parameters
    ----------
    trackid
        index of the track.
    pos_x
        x position of the step.
    pos_y
        y position of the step.
    pos_z
        z position of the step.
    dist
        distance to the detector surface. Can be `None` in which case all steps are treated as being in the "bulk".
    surf_cut
        Size of the surface region (in mm), if `None` no selection is applied (default).
    threshold
        Distance threshold in mm to combine steps in the bulk.
    threshold_surf
        Distance threshold in mm to combine steps in the surface.

    Returns
    -------
    Array of the run lengths of each cluster within a hit.
    """
    # type conversions
    if isinstance(pos_x, VectorOfVectors):
        pos_x = pos_x.view_as("ak")

    if isinstance(pos_y, VectorOfVectors):
        pos_y = pos_y.view_as("ak")

    if isinstance(pos_z, VectorOfVectors):
        pos_z = pos_z.view_as("ak")

    if isinstance(trackid, VectorOfVectors):
        trackid = trackid.view_as("ak")

    if isinstance(dist, VectorOfVectors):
        dist = dist.view_as("ak")

    pos = np.vstack(
        [
            ak.flatten(pos_x).to_numpy(),
            ak.flatten(pos_y).to_numpy(),
            ak.flatten(pos_z).to_numpy(),
        ]
    ).T

    indices_flat = cluster_by_distance_numba(
        ak.flatten(ak.local_index(trackid)).to_numpy(),
        ak.flatten(trackid).to_numpy(),
        pos,
        dist_to_surf=ak.flatten(dist).to_numpy() if dist is not None else dist,
        surf_cut=surf_cut,
        threshold=threshold,
        threshold_surf=threshold_surf,
    )

    # reshape into being event oriented
    indices = ak.unflatten(indices_flat, ak.num(ak.local_index(trackid)))

    # number of steps per cluster
    counts = ak.run_lengths(indices)

    return VectorOfVectors(counts)


@numba.njit
def cluster_by_distance_numba(
    local_index: np.ndarray,
    trackid: np.ndarray,
    pos: np.ndarray,
    dist_to_surf: np.ndarray | None,
    surf_cut: float | None = None,
    threshold: float = 0.1,
    threshold_surf: float | None = None,
) -> np.ndarray:
    """Cluster steps by the distance between points in the same track.

    This function gives the basic numerical calculations for
    :func:`cluster_by_step_length`.

    Parameters
    ----------
    local_index
        1D array of the local index within each hit (step group)
    trackid
        1D array of index of the track
    pos
        `(n,3)` size array of the positions
    dist_to_surf
        1D array of the distance to the detector surface. Can be `None` in which case all steps are treated as being in the bulk.
    surf_cut
        Size of the surface region (in mm), if `None` no selection is applied.
    threshold
        Distance threshold in mm to combine steps in the bulk.
    threshold_surf
        Distance threshold in mm to combine steps in the surface.

    Returns
    -------
    np.ndarray
        1D array of cluster indices
    """

    def _dist(a, b):
        return np.sqrt(np.sum((a - b) ** 2))

    n = len(local_index)
    out = np.zeros(n, dtype=numba.int32)

    trackid_prev = -1
    pos_prev = np.zeros(3, dtype=numba.float64)
    cluster_idx = -1
    is_surf_prev = False

    for idx in range(n):
        # consider a surface and a bulk region
        if dist_to_surf is not None:
            thr = threshold if dist_to_surf[idx] > surf_cut else threshold_surf

            new_cluster = (
                (trackid[idx] != trackid_prev)
                or (is_surf_prev and (dist_to_surf[idx] > surf_cut))
                or ((not is_surf_prev) and (dist_to_surf[idx] < surf_cut))
                or (_dist(pos[idx, :], pos_prev) > thr)
            )
        # basic clustering without split into surface / bulk
        else:
            thr = threshold
            new_cluster = (trackid[idx] != trackid_prev) or (_dist(pos[idx, :], pos_prev) > thr)

        # New hit, reset cluster index
        if idx == 0 or local_index[idx] == 0:
            cluster_idx = 0
            pos_prev = pos[idx]

        # either new track, moving from surface to bulk,
        # moving from bulk to surface, or stepping more than
        # the threshold. Start a new cluster.
        elif new_cluster:
            cluster_idx += 1
            pos_prev = pos[idx, :]

        out[idx] = cluster_idx

        # Update previous values
        trackid_prev = trackid[idx]
        if dist_to_surf is not None:
            is_surf_prev = dist_to_surf[idx] < surf_cut

    return out


def step_lengths(
    x_cluster: ak.Array | VectorOfVectors,
    y_cluster: ak.Array | VectorOfVectors,
    z_cluster: ak.Array | VectorOfVectors,
) -> VectorOfVectors:
    """Compute the distance between consecutive steps.

    This is based on calculating the distance between consecutive steps in the same track,
    thus the input arrays should already be clustered (have dimension 3). The output
    will have a similar shape to the input with one less entry in the outermost dimension.

    Example config (assuming that the clustered positions are obtained already):

    .. code-block:: yaml

        step_lengths: reboost.shape.cluster.step_lengths(HITS.cluster_x,HITS.cluster_y,HITS.cluster_z))

    Parameters
    ----------
    x_cluster
        The x location of each step in each cluster and event.
    y_cluster
        The y location of each step in each cluster and event.
    z_cluster
        The z location of each step in each cluster and event.

    Returns
    -------
    a `VectorOfVectors` of the step lengths in each cluster.
    """
    data = [x_cluster, y_cluster, z_cluster]

    for idx, var in enumerate(data):
        if isinstance(var, VectorOfVectors):
            data[idx] = var.view_as("ak")
        # check shape
        if data[idx].ndim != 3:
            msg = f"The input array for step lengths must be 3 dimensional not {data[idx.dim]}"
            raise ValueError(msg)

    counts = ak.num(data[0], axis=-1)
    data = np.vstack([ak.flatten(ak.flatten(var)).to_numpy() for var in data])
    dist = np.append(np.sqrt(np.sum(np.diff(data, axis=1) ** 2, axis=0)), 0)

    n_cluster = ak.num(counts, axis=-1)
    clusters = ak.unflatten(ak.Array(dist), ak.flatten(counts))

    out = ak.unflatten(clusters, n_cluster)
    return VectorOfVectors(out[:, :, :-1])



"""Routines to build the `hit` tier from the `stp` tier.

A :func:`build_hit` to parse the following configuration file:

.. code-block:: yaml

    # dictionary of objects useful for later computation. they are constructed with
    # auxiliary data (e.g. metadata). They can be accessed later as OBJECTS (all caps)
    objects:
     lmeta: LegendMetadata(ARGS.legendmetadata)
     geometry: pyg4ometry.load(ARGS.gdml)
     user_pars: dbetto.TextDB(ARGS.par)
     dataprod_pars: dbetto.TextDB(ARGS.dataprod_cycle)

    # processing chain is defined to act on a group of detectors
    processing_groups:

        # start with HPGe stuff, give it an optional name
        - name: geds

          # this is a list of included detectors (part of the processing group)
          detector_mapping:
            - output: OBJECTS.lmeta.channglmap(on=ARGS.timestamp)
             .group('system').geds
             .group('analysis.status').on
             .map('name').keys()

          # which columns we actually want to see in the output table
          outputs:
             - t0
             - evtid
             - energy
             - r90
             - drift_time

          # in this section we define objects that will be instantiated at each
          # iteration of the for loop over input tables (i.e. detectors)
          detector_objects:
             # The following assumes that the detector metadata is stored in the GDML file
             pyobj: legendhpges.make_hpge(pygeomtools.get_sensvol_metadata(OBJECTS.geometry, DETECTOR))
             phyvol: OBJECTS.geometry.physical_volume_dict[DETECTOR]
             drift_time_map: lgdo.lh5.read(DETECTOR, ARGS.dtmap_file)

          # this defines "hits", i.e. layout of the output hit table
          # group steps by time and evtid with 10us window
          hit_table_layout: reboost.shape.group_by_time(STEPS, window=10)

          # finally, the processing chain
          operations:

            t0: ak.fill_none(ak.firsts(HITS.time, axis=-1), np.nan)

            evtid: ak.fill_none(ak.firsts(HITS.__evtid, axis=-1), np.nan)

            # distance to the nplus surface in mm
            distance_to_nplus_surface_mm: reboost.hpge.distance_to_surface(
                HITS.__xloc, HITS.__yloc, HITS.__zloc,
                DETECTOR_OBJECTS.pyobj,
                DETECTOR_OBJECTS.phyvol.position.eval(),
                surface_type='nplus')

            # activness based on FCCD (no TL)
            activeness: ak.where(
                HITS.distance_to_nplus_surface_mm <
                    lmeta.hardware.detectors.germanium.diodes[DETECTOR].characterization.combined_0vbb_fccd_in_mm.value,
                0,
                1
                )

            activeness2: reboost.math.piecewise_linear(
                HITS.distance_to_nplus_surface_mm,
                PARS.tlayer[DETECTOR].start_in_mm,
                PARS.fccd_in_mm,
                )

            # summed energy of the hit accounting for activeness
            energy_raw: ak.sum(HITS.__edep * HITS.activeness, axis=-1)

            # energy with smearing
            energy: reboost.math.sample_convolve(
                scipy.stats.norm, # resolution distribution
                loc=HITS.energy_raw, # parameters of the distribution (observable to convolve)
                scale=np.sqrt(PARS.a + PARS.b * HITS.energy_raw) # another parameter
                )

            # this is going to return "run lengths" (awkward jargon)
            clusters_lengths: reboost.shape.cluster.naive(
                HITS, # can also pass the exact fields (x, y, z)
                size=1,
                units="mm"
                )

            # example of low level reduction on clusters
            energy_clustered: ak.sum(ak.unflatten(HITS.__edep, HITS.clusters_lengths), axis=-1)

            # example of using a reboost helper
            steps_clustered: reboost.shape.reduction.energy_weighted_average(HITS, HITS.clusters_lengths)

            r90: reboost.hpge.psd.r90(HITS.steps_clustered)

            drift_time: reboost.hpge.psd.drift_time(
                HITS.steps_clustered,
                DETECTOR_OBJECTS.drift_time_map
                )

        # example basic processing of steps in scintillators
        - name: lar
          detector_mapping:
           - output: scintillators

          outputs:
            - evtid
            - tot_edep_wlsr

          operations:
            tot_edep_wlsr: ak.sum(HITS[(HITS.__detuid == 0) & (HITS.__zloc < 3000)].__edep, axis=-1)

        - name: spms

          # by default, reboost looks in the steps input table for a table with the
          # same name as the current detector. This can be overridden for special processors

          detector_mapping:
           - output: OBJECTS.lmeta.channglmap(on=ARGS.timestamp)
            .group("system").spms
            .group("analysis.status").on
            .map("name").keys()
             input: lar

          outputs:
            - t0
            - evtid
            - pe_times

          detector_objects:
             meta: pygeomtools.get_sensvol_metadata(OBJECTS.geometry, DETECTOR)
             optmap_lar: lgdo.lh5.read(DETECTOR, "optmaps/pen", ARGS.optmap_path)
             optmap_pen: lgdo.lh5.read(DETECTOR, "optmaps/lar", ARGS.optmap_path)

          hit_table_layout: reboost.shape.group_by_time(STEPS, window=10)

          operations:
            pe_times_lar: reboost.spms.detected_photoelectrons(
                STEPS,
                DETECTOR_OBJECTS.optmap_lar,
                0
             )

            pe_times_pen: reboost.spms.detected_photoelectrons(
                STEPS,
                DETECTOR_OBJECTS.optmap_pen,
                1
             )

            pe_times: ak.concatenate([HITS.pe_times_lar, HITS.pe_times_pen], axis=-1)
"""

from __future__ import annotations

import copy
import logging
import time
from collections.abc import Mapping

import awkward as ak
import dbetto
from dbetto import AttrsDict
from lgdo import lh5
from lgdo.types import Struct

from reboost.iterator import GLMIterator
from reboost.profile import ProfileDict

from . import core, utils

log = logging.getLogger(__name__)


def build_hit(
    config: Mapping | str,
    args: Mapping | AttrsDict,
    stp_files: str | list[str],
    glm_files: str | list[str] | None,
    hit_files: str | list[str] | None,
    *,
    start_evtid: int = 0,
    n_evtid: int | None = None,
    in_field: str = "stp",
    out_field: str = "hit",
    buffer: int = int(5e6),
    overwrite: bool = False,
) -> None | ak.Array:
    """Build the hit tier from the remage step files.

    Parameters
    ----------
    config
        dictionary or path to YAML file containing the processing chain.
    args
        dictionary or :class:`dbetto.AttrsDict` of the global arguments.
    stp_files
        list of strings or string of the stp file path.
    glm_files
        list of strings or string of the glm file path, if `None` will be build in memory.
    hit_files
        list of strings or string of the hit file path. The `hit` file can also be `None` in which
        case the hits are returned as an `ak.Array` in memory.
    start_evtid
        first evtid to read.
    n_evtid
        number of evtid to read, if `None` read all.
    in_field
        name of the input field in the remage output.
    out_field
        name of the output field
    buffer
        buffer size for use in the `LH5Iterator`.
    overwrite
        flag to overwrite the existing output.
    """
    # extract the config file
    if isinstance(config, str):
        config = dbetto.utils.load_dict(config)

    # get the arguments
    if not isinstance(args, AttrsDict):
        args = AttrsDict(args)
    time_dict = ProfileDict()

    # get the global objects
    global_objects = AttrsDict(
        core.get_global_objects(
            expressions=config.get("objects", {}), local_dict={"ARGS": args}, time_dict=time_dict
        )
    )

    # get the input files
    files = utils.get_file_dict(stp_files=stp_files, glm_files=glm_files, hit_files=hit_files)

    output_tables = {}
    # iterate over files
    for file_idx, (stp_file, glm_file) in enumerate(zip(files.stp, files.glm)):
        msg = (
            f"... starting post processing of {stp_file} to {files.hit[file_idx]} "
            if files.hit[file_idx] is not None
            else f"... starting post processing of {stp_file}"
        )
        log.info(msg)

        # loop over processing groups
        for group_idx, proc_group in enumerate(config["processing_groups"]):
            proc_name = proc_group.get("name", "default")
            msg = f"... starting group {proc_name}"
            log.info(msg)

            if proc_name not in time_dict:
                time_dict[proc_name] = ProfileDict()

            # extract the output detectors and the mapping to input detectors
            detectors_mapping = utils.merge_dicts(
                [
                    core.get_detectors_mapping(
                        mapping["output"],
                        input_detector_name=mapping.get("input", None),
                        objects=global_objects,
                    )
                    for mapping in proc_group.get("detector_mapping")
                ]
            )

            # loop over detectors
            for in_det_idx, (in_detector, out_detectors) in enumerate(detectors_mapping.items()):
                msg = f"... processing {in_detector} (to {out_detectors})"
                log.info(msg)

                # get detector objects
                det_objects = core.get_detector_objects(
                    output_detectors=out_detectors,
                    args=args,
                    global_objects=global_objects,
                    expressions=proc_group.get("detector_objects", {}),
                    time_dict=time_dict[proc_name],
                )

                # begin iterating over the glm
                glm_it = GLMIterator(
                    glm_file,
                    stp_file,
                    lh5_group=in_detector,
                    start_row=start_evtid,
                    stp_field=in_field,
                    n_rows=n_evtid,
                    read_vertices=False,
                    buffer=buffer,
                    time_dict=time_dict[proc_name],
                )
                for stps, _, chunk_idx, _ in glm_it:
                    # converting to awkward
                    if stps is None:
                        continue

                    # convert to awkward
                    if time_dict is not None:
                        start_time = time.time()

                    ak_obj = stps.view_as("ak")

                    if time_dict is not None:
                        time_dict[proc_name].update_field("conv", start_time)

                    # produce the hit table
                    for out_det_idx, out_detector in enumerate(out_detectors):
                        # loop over the rows
                        if out_detector not in output_tables and files.hit[file_idx] is None:
                            output_tables[out_detector] = None

                        # get the attributes
                        attrs = utils.copy_units(stps)

                        if "hit_table_layout" in proc_group:
                            hit_table = core.evaluate_hit_table_layout(
                                copy.deepcopy(ak_obj),
                                expression=proc_group["hit_table_layout"],
                                time_dict=time_dict[proc_name],
                            )
                        else:
                            hit_table = copy.deepcopy(stps)

                        local_dict = {
                            "DETECTOR_OBJECTS": det_objects[out_detector],
                            "OBJECTS": global_objects,
                            "DETECTOR": out_detector,
                        }
                        # add fields
                        for field, expression in proc_group.get("operations", {}).items():
                            # evaluate the expression
                            col = core.evaluate_output_column(
                                hit_table,
                                table_name="HITS",
                                expression=expression,
                                local_dict=local_dict,
                                time_dict=time_dict[proc_name],
                                name=field,
                            )
                            hit_table.add_field(field, col)

                        # remove unwanted fields
                        if "outputs" in proc_group:
                            hit_table = core.remove_columns(
                                hit_table, outputs=proc_group["outputs"]
                            )

                        # assign units in the output table
                        hit_table = utils.assign_units(hit_table, attrs)

                        new_hit_file = (file_idx == 0) or (
                            files.hit[file_idx] != files.hit[file_idx - 1]
                        )

                        wo_mode = utils.get_wo_mode(
                            group=group_idx,
                            out_det=out_det_idx,
                            in_det=in_det_idx,
                            chunk=chunk_idx,
                            new_hit_file=new_hit_file,
                            overwrite=overwrite,
                        )

                        # now write
                        if files.hit[file_idx] is not None:
                            if time_dict is not None:
                                start_time = time.time()

                            if wo_mode != "a":
                                lh5.write(
                                    Struct({out_detector: hit_table}),
                                    out_field,
                                    files.hit[file_idx],
                                    wo_mode=wo_mode,
                                )
                            else:
                                lh5.write(
                                    hit_table,
                                    f"{out_field}/{out_detector}",
                                    files.hit[file_idx],
                                    wo_mode=wo_mode,
                                )
                            if time_dict is not None:
                                time_dict[proc_name].update_field("write", start_time)

                        else:
                            output_tables[out_detector] = core.merge(
                                hit_table, output_tables[out_detector]
                            )

    # return output table or nothing
    log.info(time_dict)

    if output_tables == {}:
        output_tables = None

    return output_tables, time_dict



from __future__ import annotations

import shutil
import uuid
from getpass import getuser
from pathlib import Path
from tempfile import gettempdir

import numba
import pytest
from legendtestdata import LegendTestData

_tmptestdir = Path(gettempdir()) / f"reboost-tests-{getuser()}-{uuid.uuid4()!s}"


@pytest.fixture(scope="session")
def tmptestdir_global():
    _tmptestdir.mkdir(exist_ok=False)
    return _tmptestdir


@pytest.fixture(scope="session")
def legendtestdata():
    ldata = LegendTestData()
    ldata.checkout("1d069f1")
    return ldata


@pytest.fixture(scope="module")
def tmptestdir(tmptestdir_global, request):
    p = tmptestdir_global / request.module.__name__
    p.mkdir(exist_ok=True)  # note: will be cleaned up globally.
    return p


def pytest_sessionfinish(exitstatus):
    if exitstatus == 0 and Path.exists(_tmptestdir):
        shutil.rmtree(_tmptestdir)


def patch_numba_for_tests():
    """Globally disable numba cache and enable bounds checking (for testing)."""
    njit_old = numba.njit

    def njit_patched(*args, **kwargs):
        kwargs.update({"cache": False, "boundscheck": True})
        return njit_old(*args, **kwargs)

    numba.njit = njit_patched


patch_numba_for_tests()



from __future__ import annotations

import awkward as ak
import pint
import pyg4ometry as pg4
import pytest
from lgdo.types import Array

from reboost import units
from reboost.units import ureg as u


def test_registry():
    assert "meter" in units.ureg


def test_pg4_to_pint():
    assert units.pg4_to_pint(1 * u.m) == 1 * u.m
    assert all(
        units.pg4_to_pint(pg4.gdml.Defines.Position("boh", 1, 1, 1, unit="m")) == (1, 1, 1) * u.m
    )

    with pytest.raises(ValueError):
        units.pg4_to_pint({"a": 1})


def test_units_convfact():
    assert units.units_convfact(1, "m") == 1

    arr = Array([1, 2, 3], attrs={"units": "m"})
    assert units.units_convfact(arr, "mm") == 1000


def test_unwrap_lgdo():
    arr = Array([1, 2, 3], attrs={"units": "m"})

    arr_view, unit = units.unwrap_lgdo(arr, "ak")
    assert isinstance(arr_view, ak.Array)
    assert isinstance(unit, pint.Unit)
    assert unit == u.m

    arr_view_2, unit = units.unwrap_lgdo(arr_view, "ak")
    assert arr_view_2 is arr_view
    assert unit is None


def test_unit_to_lh5_attr():
    assert units.unit_to_lh5_attr((1 * u.m).u) == "m"
    assert units.unit_to_lh5_attr(u.m) == "m"
    assert units.unit_to_lh5_attr(u.m / u.s) == "m/s"
    assert units.unit_to_lh5_attr(u.m**2) == "m**2"
    assert units.unit_to_lh5_attr(u.micrometer) == "µm"



from __future__ import annotations

import pytest
from scipy.interpolate import RegularGridInterpolator

from reboost.hpge.utils import HPGeScalarRZField, get_hpge_scalar_rz_field
from reboost.units import ureg as u


def test_read_hpge_map(legendtestdata):
    dt_map = get_hpge_scalar_rz_field(
        legendtestdata["lh5/hpge-drift-time-maps.lh5"],
        "V99000A",
        "drift_time",
        out_of_bounds_val=0,
    )

    assert isinstance(dt_map, HPGeScalarRZField)

    assert dt_map.r_units == u.m
    assert dt_map.z_units == u.m
    assert dt_map.φ_units == u.ns

    assert isinstance(dt_map.φ, RegularGridInterpolator)

    with pytest.raises(ValueError):
        dt_map.φ((0, -1))

    assert dt_map.φ((0, 0)) == 0
    assert dt_map.φ([(0, 0.01), (0.03, 0.03)]) == pytest.approx([135, 695])



from __future__ import annotations

import contextlib
import ctypes
import logging
import math
import multiprocessing as mp
from collections.abc import Mapping

import numpy as np
from lgdo import Histogram, lh5
from numpy.typing import NDArray

log = logging.getLogger(__name__)


class OpticalMap:
    def __init__(self, name: str, settings: Mapping[str, str], use_shmem: bool = False):
        self.settings = settings
        self.name = name
        self.use_shmem = use_shmem

        self.h_vertex = None
        self.h_hits = None
        self.h_prob = None
        self.h_prob_uncert = None

        self.binning = None

        self.__fill_hits_buf = None

        if settings is not None:
            self._single_shape = tuple(self.settings["bins"])
            self._single_stride = None

            binedge_attrs = {"units": "m"}
            bins = self.settings["bins"]
            bounds = self.settings["range_in_m"]
            self.binning = [
                Histogram.Axis(
                    None,
                    bounds[i][0],
                    bounds[i][1],
                    (bounds[i][1] - bounds[i][0]) / bins[i],
                    True,
                    binedge_attrs,
                )
                for i in range(3)
            ]

    @staticmethod
    def create_empty(name: str, settings: Mapping[str, str]) -> OpticalMap:
        om = OpticalMap(name, settings)
        om.h_vertex = om._prepare_hist()
        om.h_hits = om._prepare_hist()
        return om

    @staticmethod
    def load_from_file(lh5_file: str, group: str) -> OpticalMap:
        om = OpticalMap(group, None)

        def read_hist(name: str, fn: str, group: str = "all"):
            h = lh5.read(f"/{group}/{name}", lh5_file=fn)
            if not isinstance(h, Histogram):
                msg = f"encountered invalid optical map while reading /{group}/{name} in {fn}"
                raise RuntimeError(msg)
            return h.weights.nda, h.binning

        om.h_vertex, bin_nr_gen = read_hist("nr_gen", lh5_file, group=group)
        om.h_hits, bin_nr_det = read_hist("nr_det", lh5_file, group=group)
        om.h_prob, bin_p_det = read_hist("p_det", lh5_file, group=group)
        om.h_prob_uncert, bin_p_det_err = read_hist("p_det_err", lh5_file, group=group)

        for bins in (bin_nr_det, bin_p_det, bin_p_det_err):
            if not OpticalMap._edges_eq(bin_nr_gen, bins):
                msg = "edges of optical map histograms differ"
                raise RuntimeError(msg)

        om.binning = bin_nr_gen
        return om

    def _prepare_hist(self) -> np.ndarray:
        """Prepare an empty histogram with the parameters global to this map instance."""
        if self.use_shmem:
            assert mp.current_process().name == "MainProcess"
            a = self._mp_man.Array(ctypes.c_double, math.prod(self._single_shape))
            nda = self._nda(a)
            nda.fill(0)
        else:
            a = np.zeros(shape=self._single_shape, dtype=np.float64)
            nda = a
        stride = [s // nda.dtype.itemsize for s in nda.strides]
        if self._single_stride is None:
            self._single_stride = stride
        assert self._single_stride == stride
        return a

    def _fill_histogram(
        self,
        h: NDArray | mp.sharedctypes.SynchronizedArray,
        xyz: NDArray,
        for_hits: bool = False,
    ) -> None:
        assert xyz.shape[1] == 3
        xyz = xyz.T

        # use as much pre-allocated memory as possible.
        if self.__fill_hits_buf is None:
            self.__fill_hits_buf = np.empty(5000, np.int64)
            self.__fill_hits_pos = 0

        idx = np.zeros(xyz.shape[1], np.int64)  # bin indices for flattened array
        oor_mask = np.ones(xyz.shape[1], np.bool_)  # mask to remove out of range values
        dims = range(xyz.shape[0])
        for col, ax, s, dim in zip(xyz, self.binning, self._single_stride, dims):
            assert ax.is_range
            assert ax.closedleft
            oor_mask &= (ax.first <= col) & (col < ax.last)
            idx_s = np.floor((col.astype(np.float64) - ax.first) / ax.step).astype(np.int64)
            assert np.all(idx_s[oor_mask] < self._single_shape[dim])
            idx += s * idx_s

        idx = idx[oor_mask]
        if idx.shape[0] == 0:
            return

        if for_hits and idx.shape[0] < self.__fill_hits_buf.shape[0]:
            # special path for the typically small number of hits.
            # this circumvents a memory leak in _fill_histogram_buf when called with varying and
            # small shapes of the idx array.
            end = self.__fill_hits_pos + idx.shape[0]
            if end >= self.__fill_hits_buf.shape[0]:
                # flush the old buffer to the map, as the new data does not fit.
                self._fill_histogram_buf(h, self.__fill_hits_buf[0 : self.__fill_hits_pos])
                self.__fill_hits_pos = 0
                end = idx.shape[0]
            self.__fill_hits_buf[self.__fill_hits_pos : end] = idx
            self.__fill_hits_pos = end
        else:
            # here we assume a uniform size of idx, so that we do not hit the memory leak.
            self._fill_histogram_buf(h, idx)

    def _fill_histogram_buf(
        self,
        h: NDArray | mp.sharedctypes.SynchronizedArray,
        idx: NDArray,
    ) -> None:
        # increment bin contents
        with self._lock_nda(h)():
            np.add.at(self._nda(h).reshape(-1), idx, 1)

    def _nda(self, h: NDArray | mp.sharedctypes.SynchronizedArray) -> NDArray:
        if not self.use_shmem:
            return h
        return np.ndarray(self._single_shape, dtype=np.float64, buffer=h.get_obj())

    def _lock_nda(self, h: NDArray | mp.sharedctypes.SynchronizedArray):
        if not self.use_shmem:
            return contextlib.nullcontext
        return h.get_lock

    def _mp_preinit(self, mp_man: mp.context.BaseContext, vertex: bool) -> None:
        self._mp_man = mp_man
        if self.h_vertex is None and vertex:
            self.h_vertex = self._prepare_hist()
        if self.h_hits is None:
            self.h_hits = self._prepare_hist()

    def fill_vertex(self, loc: NDArray) -> None:
        """Fill map with a chunk of hit coordinates."""
        if self.h_vertex is None:
            self.h_vertex = self._prepare_hist()
        self._fill_histogram(self.h_vertex, loc)

    def fill_hits(self, loc: NDArray) -> None:
        """Fill map with a chunk of hit coordinates.

        .. note::

            For performance reasons, this function is buffered and does not
            directly write to the map array. Use :meth:`.fill_hits_flush` to
            flush the remaining hits in the buffer to this map.
        """
        if self.h_hits is None:
            self.h_hits = self._prepare_hist()
        self._fill_histogram(self.h_hits, loc, for_hits=True)

    def fill_hits_flush(self) -> None:
        """Commit all remaining hit coordinates in the buffer."""
        if self.h_hits is None or self.__fill_hits_pos <= 0:
            return
        self._fill_histogram_buf(self.h_hits, self.__fill_hits_buf[0 : self.__fill_hits_pos])
        self.__fill_hits_buf = None

    def _divide_hist(self, h1: NDArray, h2: NDArray) -> tuple[NDArray, NDArray]:
        """Calculate the ratio (and its standard error) from two histograms."""
        h1 = self._nda(h1)
        h2 = self._nda(h2)

        ratio_0 = self._prepare_hist()
        ratio_err_0 = self._prepare_hist()
        ratio, ratio_err = self._nda(ratio_0), self._nda(ratio_err_0)

        ratio[:] = np.divide(h1, h2, where=(h2 != 0))
        ratio[h2 == 0] = -1  # -1 denotes no statistics.

        if np.any(ratio > 1):
            msg = "encountered cell(s) with more hits than primaries"
            raise RuntimeError(msg)

        # compute uncertainty according to Bernoulli statistics.
        # TODO: this does not make sense for ratio==1
        ratio_err[h2 != 0] = np.sqrt((ratio[h2 != 0]) * (1 - ratio[h2 != 0]) / h2[h2 != 0])
        ratio_err[h2 == 0] = -1  # -1 denotes no statistics.

        return ratio_0, ratio_err_0

    def create_probability(self) -> None:
        """Compute probability map (and map uncertainty) from vertex and hit map."""
        self.h_prob, self.h_prob_uncert = self._divide_hist(self.h_hits, self.h_vertex)

    def write_lh5(self, lh5_file: str, group: str = "all", wo_mode: str = "write_safe") -> None:
        """Write this map to a LH5 file."""
        if wo_mode not in ("write_safe", "overwrite_file"):
            msg = f"invalid wo_mode {wo_mode} for optical map"
            raise ValueError(msg)

        def write_hist(h: NDArray, name: str, fn: str, group: str, wo_mode: str):
            lh5.write(
                Histogram(self._nda(h), self.binning),
                name,
                fn,
                group=group,
                wo_mode=wo_mode,
            )

        # only use the passed wo_mode for the first file.
        write_hist(self.h_vertex, "nr_gen", lh5_file, group, wo_mode)
        write_hist(self.h_hits, "nr_det", lh5_file, group, "write_safe")
        write_hist(self.h_prob, "p_det", lh5_file, group, "write_safe")
        write_hist(self.h_prob_uncert, "p_det_err", lh5_file, group, "write_safe")

    def get_settings(self) -> dict:
        """Get the binning settings that were used to create this optical map instance."""
        if self.settings is not None:
            return self.settings

        range_in_m = []
        bins = []
        for b in self.binning:
            if not b.is_range:
                msg = "cannot get binning settings for variable binning map"
                raise RuntimeError(msg)
            if b.get_binedgeattrs().get("units") != "m":
                msg = "invalid units. can only work with optical maps in meter"
                raise RuntimeError(msg)
            range_in_m.append([b.first, b.last])
            bins.append(b.nbins)

        return {"range_in_m": np.array(range_in_m), "bins": np.array(bins)}

    def check_histograms(self, include_prefix: bool = False) -> None:
        log_prefix = "" if not include_prefix else self.name + " - "

        def _warn(fmt: str, *args):
            log.warning("%s" + fmt, log_prefix, *args)  # noqa: G003

        h_vertex = self._nda(self.h_vertex)
        h_prob = self._nda(self.h_prob)
        h_prob_uncert = self._nda(self.h_prob_uncert)

        ncells = h_vertex.shape[0] * h_vertex.shape[1] * h_vertex.shape[2]

        missing_v = np.sum(h_vertex <= 0)  # bins without vertices.
        if missing_v > 0:
            _warn("%d missing_v %.2f %%", missing_v, missing_v / ncells * 100)

        missing_p = np.sum(h_prob <= 0)  # bins without hist.
        if missing_p > 0:
            _warn("%d missing_p %.2f %%", missing_p, missing_p / ncells * 100)

        non_phys = np.sum(h_prob > 1)  # non-physical events with probability > 1.
        if non_phys > 0:
            _warn(
                "%d voxels (%.2f %%) with non-physical probability (p>1)",
                non_phys,
                non_phys / ncells * 100,
            )

        # warnings on insufficient statistics.
        large_error = np.sum(h_prob_uncert > 0.01 * h_prob)
        if large_error > 0:
            _warn(
                "%d voxels (%.2f %%) with large relative statistical uncertainty (> 1 %%)",
                large_error,
                large_error / ncells * 100,
            )

        primaries_low_stats_th = 100
        low_stat_zero = np.sum((h_vertex < primaries_low_stats_th) & (h_prob == 0))
        if low_stat_zero > 0:
            _warn(
                "%d voxels (%.2f %%) with non reliable probability estimate (p=0 and primaries < %d)",
                low_stat_zero,
                low_stat_zero / ncells * 100,
                primaries_low_stats_th,
            )
        low_stat_one = np.sum((h_vertex < primaries_low_stats_th) & (h_prob == 1))
        if low_stat_one > 0:
            _warn(
                "%d voxels (%.2f %%) with non reliable probability estimate (p=1 and primaries < %d)",
                low_stat_one,
                low_stat_one / ncells * 100,
                primaries_low_stats_th,
            )

    @staticmethod
    def _edges_eq(
        e1: tuple[NDArray] | tuple[Histogram.Axis], e2: tuple[NDArray] | tuple[Histogram.Axis]
    ) -> bool:
        """Compare edge-tuples for two histograms."""
        if isinstance(e1[0], Histogram.Axis):
            e1 = tuple([b.edges for b in e1])
        if isinstance(e2[0], Histogram.Axis):
            e2 = tuple([b.edges for b in e2])
        assert all(isinstance(b, np.ndarray) for b in e1)
        assert all(isinstance(b, np.ndarray) for b in e2)

        return len(e1) == len(e2) and all(np.all(x1 == x2) for x1, x2 in zip(e1, e2))



from __future__ import annotations

import argparse
import json
import logging
from pathlib import Path

from ..log_utils import setup_log
from ..utils import _check_input_file, _check_output_file

log = logging.getLogger(__name__)


def optical_cli() -> None:
    parser = argparse.ArgumentParser(
        prog="reboost-optical",
        description="%(prog)s command line interface",
    )

    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=0,
        help="""Increase the program verbosity""",
    )

    parser.add_argument(
        "--bufsize",
        action="store",
        type=int,
        default=int(5e6),
        help="""Row count for input table buffering (only used if applicable). default: %(default)e""",
    )

    subparsers = parser.add_subparsers(dest="command", required=True)

    # STEP 1: build evt file from hit tier
    evt_parser = subparsers.add_parser("evt", help="build evt file from remage hit file")
    evt_parser.add_argument(
        "--detectors",
        help="file that contains a list of detector ids that are part of the input file",
        required=True,
    )
    evt_parser.add_argument("input", help="input hit LH5 file", metavar="INPUT_HIT")
    evt_parser.add_argument("output", help="output evt LH5 file", metavar="OUTPUT_EVT")

    # STEP 2a: build map file from evt tier
    map_parser = subparsers.add_parser("createmap", help="build optical map from evt file(s)")
    map_parser.add_argument(
        "--settings",
        action="store",
        help="""Select a config file for binning.""",
        required=True,
    )
    map_parser.add_argument(
        "--detectors",
        help="file that contains a list of detector ids that will be produced as additional output maps.",
    )
    map_parser.add_argument(
        "--n-procs",
        "-N",
        type=int,
        default=1,
        help="number of worker processes to use. default: %(default)e",
    )
    map_parser.add_argument(
        "--check",
        action="store_true",
        help="""Check map statistics after creation. default: %(default)s""",
    )
    map_parser.add_argument("input", help="input evt LH5 file", metavar="INPUT_EVT", nargs="+")
    map_parser.add_argument("output", help="output map LH5 file", metavar="OUTPUT_MAP")

    # STEP 2b: view maps
    mapview_parser = subparsers.add_parser("viewmap", help="view optical map")
    mapview_parser.add_argument("input", help="input map LH5 file", metavar="INPUT_MAP")
    mapview_parser.add_argument(
        "--channel",
        action="store",
        default="all",
        help="default: %(default)s",
    )
    mapview_parser.add_argument(
        "--hist",
        choices=("nr_gen", "nr_det", "p_det", "p_det_err", "p_det_err_rel"),
        action="store",
        default="p_det",
        help="select optical map histogram to show. default: %(default)s",
    )
    mapview_parser.add_argument(
        "--divide",
        action="store",
        help="default: none",
    )
    mapview_parser.add_argument(
        "--min",
        default=1e-4,
        type=(lambda s: s if s == "auto" else float(s)),
        help="colormap min value. default: %(default)e",
    )
    mapview_parser.add_argument(
        "--max",
        default=1e-2,
        type=(lambda s: s if s == "auto" else float(s)),
        help="colormap max value. default: %(default)e",
    )
    mapview_parser.add_argument("--title", help="title of figure. default: stem of filename")

    # STEP 2c: merge maps
    mapmerge_parser = subparsers.add_parser("mergemap", help="merge optical maps")
    mapmerge_parser.add_argument(
        "input", help="input map LH5 files", metavar="INPUT_MAP", nargs="+"
    )
    mapmerge_parser.add_argument("output", help="output map LH5 file", metavar="OUTPUT_MAP")
    mapmerge_parser.add_argument(
        "--settings",
        action="store",
        help="""Select a config file for binning.""",
        required=True,
    )
    mapmerge_parser.add_argument(
        "--n-procs",
        "-N",
        type=int,
        default=1,
        help="number of worker processes to use. default: %(default)e",
    )
    mapmerge_parser.add_argument(
        "--check",
        action="store_true",
        help="""Check map statistics after creation. default: %(default)s""",
    )

    # STEP 2d: check map
    checkmap_parser = subparsers.add_parser("checkmap", help="check optical maps")
    checkmap_parser.add_argument("input", help="input map LH5 file", metavar="INPUT_MAP")

    # STEP 3: convolve with hits from non-optical simulations
    convolve_parser = subparsers.add_parser(
        "convolve", help="convolve non-optical hits with optical map"
    )
    convolve_parser.add_argument(
        "--material",
        action="store",
        choices=("lar", "pen", "fib"),
        default="lar",
        help="default: %(default)s",
    )
    convolve_parser.add_argument(
        "--map",
        action="store",
        required=True,
        metavar="INPUT_MAP",
        help="input map LH5 file",
    )
    convolve_parser.add_argument(
        "--edep",
        action="store",
        required=True,
        metavar="INPUT_EDEP",
        help="input non-optical LH5 hit file",
    )
    convolve_parser.add_argument(
        "--edep-lgdo",
        action="store",
        required=True,
        metavar="LGDO_PATH",
        help="path to LGDO inside non-optical LH5 hit file (e.g. /stp/detXX)",
    )
    convolve_parser.add_argument(
        "--dist-mode",
        action="store",
        default="multinomial+no-fano",
    )
    convolve_parser.add_argument("--output", help="output hit LH5 file", metavar="OUTPUT_HIT")

    # STEP X: rebin maps
    rebin_parser = subparsers.add_parser("rebin", help="rebin optical maps")
    rebin_parser.add_argument("input", help="input map LH5 files", metavar="INPUT_MAP")
    rebin_parser.add_argument("output", help="output map LH5 file", metavar="OUTPUT_MAP")
    rebin_parser.add_argument("--factor", type=int, help="integer scale-down factor")

    args = parser.parse_args()

    log_level = (None, logging.INFO, logging.DEBUG)[min(args.verbose, 2)]
    setup_log(log_level)

    # STEP 1: build evt file from hit tier
    if args.command == "evt":
        from reboost.optmap.evt import build_optmap_evt

        _check_input_file(parser, args.detectors)
        _check_input_file(parser, args.input)
        _check_output_file(parser, args.output)

        # load detector ids from a JSON array
        with Path.open(Path(args.detectors)) as detectors_f:
            detectors = json.load(detectors_f)

        build_optmap_evt(args.input, args.output, detectors, args.bufsize)

    # STEP 2a: build map file from evt tier
    if args.command == "createmap":
        from reboost.optmap.create import create_optical_maps

        _check_input_file(parser, args.input)
        _check_output_file(parser, args.output)

        # load settings for binning from config file.
        _check_input_file(parser, args.input, "settings")
        with Path.open(Path(args.settings)) as settings_f:
            settings = json.load(settings_f)

        chfilter = ()
        if args.detectors is not None:
            # load detector ids from a JSON array
            with Path.open(Path(args.detectors)) as detectors_f:
                chfilter = json.load(detectors_f)

        create_optical_maps(
            args.input,
            settings,
            args.bufsize,
            chfilter=chfilter,
            output_lh5_fn=args.output,
            check_after_create=args.check,
            n_procs=args.n_procs,
        )

    # STEP 2b: view maps
    if args.command == "viewmap":
        from reboost.optmap.mapview import view_optmap

        _check_input_file(parser, args.input)
        if args.divide is not None:
            _check_input_file(parser, args.divide)
        view_optmap(
            args.input,
            args.channel,
            args.divide,
            cmap_min=args.min,
            cmap_max=args.max,
            title=args.title,
            histogram_choice=args.hist,
        )

    # STEP 2c: merge maps
    if args.command == "mergemap":
        from reboost.optmap.create import merge_optical_maps

        # load settings for binning from config file.
        _check_input_file(parser, args.input, "settings")
        with Path.open(Path(args.settings)) as settings_f:
            settings = json.load(settings_f)

        _check_input_file(parser, args.input)
        _check_output_file(parser, args.output)
        merge_optical_maps(
            args.input, args.output, settings, check_after_create=args.check, n_procs=args.n_procs
        )

    # STEP 2d: check maps
    if args.command == "checkmap":
        from reboost.optmap.create import check_optical_map

        _check_input_file(parser, args.input)
        check_optical_map(args.input)

    # STEP 3: convolve with hits from non-optical simulations
    if args.command == "convolve":
        from reboost.optmap.convolve import convolve

        _check_input_file(parser, [args.map, args.edep])
        _check_output_file(parser, args.output, optional=True)
        convolve(
            args.map,
            args.edep,
            args.edep_lgdo,
            args.material,
            args.output,
            args.bufsize,
            dist_mode=args.dist_mode,
        )

    # STEP X: rebin maps
    if args.command == "rebin":
        from reboost.optmap.create import rebin_optical_maps

        _check_input_file(parser, args.input)
        _check_output_file(parser, args.output)
        rebin_optical_maps(args.input, args.output, args.factor)



from __future__ import annotations

import argparse
import logging

import dbetto

from reboost.build_glm import build_glm
from reboost.build_hit import build_hit
from reboost.utils import _check_input_file, _check_output_file, get_file_list

from .log_utils import setup_log

log = logging.getLogger(__name__)


def cli(args=None) -> None:
    parser = argparse.ArgumentParser(
        prog="reboost",
        description="%(prog)s command line interface",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=1,
        help="""Increase the program verbosity""",
    )

    subparsers = parser.add_subparsers(dest="command", required=True)

    # glm parser
    glm_parser = subparsers.add_parser("build-glm", help="build glm file from remage stp file")

    glm_parser.add_argument(
        "--stp-file",
        "-s",
        required=True,
        type=str,
        help="Path to the stp file, if multithreaded this will be appended with _t$idx.",
    )
    glm_parser.add_argument(
        "--glm-file",
        "-g",
        required=True,
        type=str,
        help="Path to the glm file, if multithreaded this will be appended with _t$idx. ",
    )

    # optional args
    glm_parser.add_argument(
        "--out-table-name", "-n", type=str, default="glm", help="Output table name."
    )
    glm_parser.add_argument("--id-name", "-i", type=str, default="g4_evtid", help="ID column name.")
    glm_parser.add_argument(
        "--evtid-buffer", "-e", type=int, default=int(1e7), help="event id buffer size."
    )
    glm_parser.add_argument(
        "--stp-buffer", "-b", type=int, default=int(1e7), help="stp buffer size."
    )
    glm_parser.add_argument(
        "--overwrite", "-w", action="store_true", help="Overwrite the input file if it exists."
    )
    glm_parser.add_argument(
        "--threads",
        "-t",
        required=False,
        default=None,
        type=int,
        help="Number of threads used for remage",
    )

    # hit parser
    hit_parser = subparsers.add_parser("build-hit", help="build hit file from remage stp file")

    hit_parser.add_argument(
        "--config", type=str, required=True, help="Path to the configuration file ."
    )
    hit_parser.add_argument("--args", type=str, required=True, help="Path to args file.")
    hit_parser.add_argument(
        "--stp-file",
        type=str,
        required=True,
        help="stp file to process, if multithreaded this will be appended with _t$idx",
    )
    hit_parser.add_argument(
        "--glm-file",
        type=str,
        required=True,
        help="glm file to process, if multithreaded this will be appended with _t$idx",
    )
    hit_parser.add_argument(
        "--hit-file",
        type=str,
        required=True,
        help="hit file to produce, if multithreaded this will be appended with _t$idx",
    )

    # optional args
    hit_parser.add_argument("--start-evtid", type=int, default=0, help="Start event id.")
    hit_parser.add_argument(
        "--n-evtid", type=int, default=None, help="Number of event id to process."
    )
    hit_parser.add_argument("--in-field", type=str, default="stp", help="Input field name.")
    hit_parser.add_argument("--out-field", type=str, default="hit", help="Output field name.")
    hit_parser.add_argument("--buffer", type=int, default=int(5e6), help="Buffer size.")

    hit_parser.add_argument(
        "--overwrite", "-w", action="store_true", help="Overwrite the input file if it exists."
    )
    hit_parser.add_argument(
        "--threads",
        "-t",
        required=False,
        default=None,
        type=int,
        help="Number of threads used for remage",
    )

    args = parser.parse_args(args)

    log_level = (None, logging.INFO, logging.DEBUG)[min(args.verbose, 2)]
    setup_log(log_level)

    if args.command == "build-glm":
        # catch some cases
        glm_files = get_file_list(args.glm_file, threads=args.threads)
        stp_files = get_file_list(args.stp_file, threads=args.threads)

        _check_input_file(parser, stp_files)

        if args.overwrite is False:
            _check_output_file(parser, glm_files)

        msg = "Running build_glm with arguments: \n"
        msg += f"    glm file:       {glm_files}\n"
        msg += f"    stp file:       {stp_files}\n"
        msg += f"    out_table_name: {args.out_table_name}\n"
        msg += f"    evtid_name:     {args.id_name}\n"
        msg += f"    evtid_buffer:   {args.evtid_buffer}\n"
        msg += f"    stp_buffer:     {args.stp_buffer}"

        log.info(msg)

        build_glm(
            stp_files,
            glm_files,
            out_table_name=args.out_table_name,
            id_name=args.id_name,
            evtid_buffer=args.evtid_buffer,
            stp_buffer=args.stp_buffer,
        )

    elif args.command == "build-hit":
        glm_files = get_file_list(args.glm_file, threads=args.threads)
        stp_files = get_file_list(args.stp_file, threads=args.threads)
        hit_files = get_file_list(args.hit_file, threads=args.threads)

        _check_input_file(parser, stp_files)
        _check_input_file(parser, glm_files)

        if args.overwrite is False:
            _check_output_file(parser, hit_files)

        msg = "Running build_hit with arguments: \n"
        msg += f"    config:         {args.config}\n"
        msg += f"    args:           {args.args}\n"
        msg += f"    glm files:      {glm_files}\n"
        msg += f"    stp files:      {stp_files}\n"
        msg += f"    hit files:      {hit_files}\n"
        msg += f"    start_evtid:    {args.start_evtid}\n"
        msg += f"    n_evtid:        {args.n_evtid}\n"
        msg += f"    in_field:       {args.in_field}\n"
        msg += f"    out_field:      {args.out_field}\n"
        msg += f"    buffer:         {args.buffer}"
        msg += f"    overwrite:      {args.overwrite}"

        log.info(msg)

        build_hit(
            config=args.config,
            args=dbetto.AttrsDict(dbetto.utils.load_dict(args.args)),
            stp_files=stp_files,
            glm_files=glm_files,
            hit_files=hit_files,
            start_evtid=args.start_evtid,
            n_evtid=args.n_evtid,
            in_field=args.in_field,
            out_field=args.out_field,
            buffer=args.buffer,
            overwrite=args.overwrite,
        )



from __future__ import annotations

import importlib
import json
from pathlib import Path

import pytest
import yaml
from lgdo.types import Array, Table

import reboost
from reboost.shape import group
from reboost.utils import (
    assign_units,
    copy_units,
    get_file_dict,
    get_function_string,
    get_wo_mode,
    merge_dicts,
)


def test_search_string():
    assert reboost.utils._search_string("func.module(abc)") == ["func.module"]
    assert reboost.utils._search_string("f.c.a(abc())") == ["f.c.a", "abc"]
    assert reboost.utils._search_string("f.c.a(a(),abc)") == ["f.c.a", "a"]
    assert reboost.utils._search_string("f.c.a(a(),abc).method()") == ["f.c.a", "a", "method"]


def test_get_function_string():
    # test that the expression looks correct and the packages are properly imported

    # basic numpy
    expression = "numpy.sum(energy)"
    energy = [1, 2, 3]
    func_string, globals_dict = get_function_string(expression)
    assert func_string == "np.sum(energy)"
    assert eval(func_string, {"energy": energy}, globals_dict) == 6

    # check the import was correct
    numpy = importlib.import_module("numpy")
    assert list(globals_dict.keys()) == ["np"]
    assert globals_dict["np"] == numpy

    # now try instead awkward
    expression = "awkward.num(energy)"
    energy = [[1, 2], [1, 2, 3], [1, 2, 3, 4]]
    func_string, globals_dict = get_function_string(expression)
    assert func_string == "ak.num(energy)"
    assert list(eval(func_string, {"energy": energy}, globals_dict)) == [2, 3, 4]

    awkward = importlib.import_module("awkward")
    assert list(globals_dict.keys()) == ["ak"]
    assert globals_dict["ak"] == awkward

    # try a chain operation
    expression = "ak.sum(np.array(energy))"
    func_string, globals_dict = get_function_string(expression)
    assert func_string == expression
    assert eval(func_string, {"energy": [1, 2, 3]}, globals_dict) == 6
    assert sorted(globals_dict.keys()) == ["ak", "np"]
    assert globals_dict["ak"] == awkward
    assert globals_dict["np"] == numpy

    # try a reboost package

    expression = "reboost.math.functions.piecewise_linear_activeness(distances,fccd=1,dlf=0.5)"
    func_string, globals_dict = get_function_string(expression)

    assert func_string == expression
    assert list(globals_dict.keys()) == ["reboost"]

    distances = awkward.Array([[0.2], [2], [0.6]])

    # just check it runs
    assert eval(func_string, {"distances": distances}, globals_dict).view_as("ak")[0][0] == 0
    assert eval(func_string, {"distances": distances}, globals_dict).view_as("ak")[1][0] == 1
    assert eval(func_string, {"distances": distances}, globals_dict).view_as("ak")[2][
        0
    ] == pytest.approx(0.2)

    # try a more compliated expression
    expression = (
        "legendhpges.make_hpge(pygeomtools.get_sensvol"
        "_metadata(OBJECTS.geometry, DETECTOR),registry = OBJECT.reg,name=ARGS.name)"
    )

    func_string, globals_dict = get_function_string(expression)

    assert list(globals_dict.keys()) == ["legendhpges", "pygeomtools"]


def test_merge_dicts():
    assert merge_dicts([{"a": [1, 2, 3], "b": [2]}, {"a": [4, 5, 6], "c": [2]}]) == {
        "a": [1, 2, 3, 4, 5, 6],
        "b": [2],
        "c": [2],
    }


@pytest.fixture
def test_save_dict(tmp_path):
    data = {"a": 1, "b": {"c": 1}}

    # Dumping JSON
    with Path(tmp_path / "data.json").open("w") as json_file:
        json.dump(data, json_file, indent=2)

    # Dumping YAML
    with Path(tmp_path / "data.yaml").open("w") as yaml_file:
        yaml.dump(data, yaml_file, default_flow_style=False)

    # make a csv too
    with Path(tmp_path / "data.csv").open("w") as yaml_file:
        yaml.dump(data, yaml_file, default_flow_style=False)

    return tmp_path


def test_wo_mode():
    assert get_wo_mode(0, 0, 0, 0, True, overwrite=True) == "of"
    assert get_wo_mode(0, 0, 0, 0, True, overwrite=False) == "w"
    assert get_wo_mode(0, 0, 1, 0, True, overwrite=False) == "ac"
    assert get_wo_mode(0, 0, 1, 1, True, overwrite=False) == "a"


def test_get_files_dict():
    # simplest case - all are str
    files = get_file_dict(stp_files="stp.lh5", hit_files="hit.lh5", glm_files="glm.lh5")

    assert files.stp == ["stp.lh5"]
    assert files.hit == ["hit.lh5"]
    assert files.glm == ["glm.lh5"]

    # also support all being lists
    files = get_file_dict(
        stp_files=["stp1.lh5", "stp2.lh5"],
        hit_files=["hit1.lh5", "hit2.lh5"],
        glm_files=["glm1.lh5", "glm2.lh5"],
    )

    assert files.stp == ["stp1.lh5", "stp2.lh5"]
    assert files.hit == ["hit1.lh5", "hit2.lh5"]
    assert files.glm == ["glm1.lh5", "glm2.lh5"]

    # hit file being None should be supported

    files = get_file_dict(
        stp_files=["stp1.lh5", "stp2.lh5"],
        hit_files=None,
        glm_files=["glm1.lh5", "glm2.lh5"],
    )

    assert files.stp == ["stp1.lh5", "stp2.lh5"]
    assert files.hit == [None, None]
    assert files.glm == ["glm1.lh5", "glm2.lh5"]

    files = get_file_dict(
        stp_files="stp.lh5",
        hit_files=None,
        glm_files="glm.lh5",
    )

    assert files.stp == ["stp.lh5"]
    assert files.hit == [None]
    assert files.glm == ["glm.lh5"]

    # glm file can be None
    files = get_file_dict(
        stp_files=["stp1.lh5", "stp2.lh5"], hit_files=["hit1.lh5", "hit2.lh5"], glm_files=None
    )

    assert files.stp == ["stp1.lh5", "stp2.lh5"]
    assert files.hit == ["hit1.lh5", "hit2.lh5"]
    assert files.glm == [None, None]

    # and hit file can be a string
    files = get_file_dict(
        stp_files=["stp1.lh5", "stp2.lh5"], hit_files="hit.lh5", glm_files=["glm1.lh5", "glm2.lh5"]
    )
    assert files.stp == ["stp1.lh5", "stp2.lh5"]
    assert files.hit == ["hit.lh5", "hit.lh5"]
    assert files.glm == ["glm1.lh5", "glm2.lh5"]

    # list of one hit file should also work
    files = get_file_dict(
        stp_files=["stp1.lh5", "stp2.lh5"],
        hit_files=["hit.lh5"],
        glm_files=["glm1.lh5", "glm2.lh5"],
    )
    assert files.stp == ["stp1.lh5", "stp2.lh5"]
    assert files.hit == ["hit.lh5", "hit.lh5"]
    assert files.glm == ["glm1.lh5", "glm2.lh5"]


def test_units():
    table = Table({"a": Array([1, 2, 3]), "b": Array([4, 5, 6]), "evtid": Array([0, 0, 1])})

    table.a.attrs = {"datatype": "array<1>{real}", "units": "ns"}
    table.b.attrs = {"datatype": "array<1>{real}", "units": "keV"}

    units = copy_units(table)

    assert units["a"] == "ns"
    assert units["b"] == "keV"
    reshaped = group.group_by_evtid(table.view_as("ak"))

    # also add an array field
    units["c"] = "keV"
    reshaped["c"] = Array([1, 2])
    reshaped = assign_units(reshaped, units)

    assert reshaped.a.flattened_data.attrs["units"] == "ns"
    assert reshaped.b.flattened_data.attrs["units"] == "keV"
    assert reshaped.c.attrs["units"] == "keV"


def test_get_channels():
    pass



from __future__ import annotations

import logging
import re

import awkward as ak
from lgdo import Table, lh5

from reboost.shape import group

log = logging.getLogger(__name__)


def build_tcm(
    hit_file: str,
    out_file: str,
    channels: list[str],
    time_name: str = "t0",
    idx_name: str = "global_evtid",
    time_window_in_us: float = 10,
) -> None:
    """Build the (Time Coincidence Map) TCM from the hit tier.

    Parameters
    ----------
    hit_file
        path to hit tier file.
    out_file
        output path for tcm.
    channels
        list of channel names to include.
    time_name
        name of the hit tier field used for time grouping.
    idx_name
        name of the hit tier field used for index grouping.
    time_window_in_us
        time window used to define the grouping.
    """
    hash_func = r"\d+"

    msg = "start building time-coincidence map"
    log.info(msg)

    chan_ids = [re.search(hash_func, channel).group() for channel in channels]

    hit_data = []
    for channel in channels:
        hit_data.append(
            lh5.read(f"{channel}/hit", hit_file, field_mask=[idx_name, time_name]).view_as("ak")
        )
    tcm = get_tcm_from_ak(
        hit_data, chan_ids, window=time_window_in_us, time_name=time_name, idx_name=idx_name
    )

    if tcm is not None:
        lh5.write(tcm, "tcm", out_file, wo_mode="of")


def get_tcm_from_ak(
    hit_data: list[ak.Array],
    channels: list[int],
    *,
    window: float = 10,
    time_name: str = "t0",
    idx_name: str = "global_evtid",
) -> Table:
    """Builds a time-coincidence map from a hit of hit data Tables.

    - build an ak.Array of the data merging channels with fields base on "time_name", and "idx_name" and adding a field `rawid` from the channel idx, also add the row (`hit_idx`)
    - sorts this array by "idx_name" then "time_name" fields
    - group by "idx_name" and "time_name" based on the window parameter

    Parameters
    ----------
    hit_data
        list of hit tier data for each channel
    channels
        list of channel indices
    window
        time window for selecting coincidences (in us)
    time_name
        name of the field for time information
    idx_name
        name of the decay index field

    Returns
    -------
    an LGDO.VectorOfVectors containing the time-coincidence map
    """
    # build ak_obj for sorting
    sort_objs = []

    for ch_idx, data_tmp in zip(channels, hit_data):
        obj_tmp = ak.copy(data_tmp)
        obj_tmp = obj_tmp[[time_name, idx_name]]
        hit_idx = ak.local_index(obj_tmp)

        obj_tmp = ak.with_field(obj_tmp, hit_idx, "array_idx")

        obj_tmp["array_id"] = int(ch_idx)
        sort_objs.append(obj_tmp)

    obj_tot = ak.concatenate(sort_objs)

    return group.group_by_time(
        obj_tot,
        time_name=time_name,
        evtid_name=idx_name,
        window=window,
        fields=["array_id", "array_idx"],
    )



from __future__ import annotations

import logging
from collections.abc import Iterable
from pathlib import Path

import numpy as np
from lgdo import lh5
from lgdo.lh5 import LH5Iterator
from lgdo.types import Table

log = logging.getLogger(__name__)

EVT_TABLE_NAME = "optmap_evt"


def build_optmap_evt(
    lh5_in_file: str, lh5_out_file: str, detectors: Iterable[str | int], buffer_len: int = int(5e6)
) -> None:
    """Create a faster map for lookup of the hits in each detector, for each primary event."""
    log.info("reading file %s", lh5_in_file)

    lh5_out_file = Path(lh5_out_file)
    lh5_out_file_tmp = lh5_out_file.with_stem(".evt-tmp." + lh5_out_file.stem)
    if lh5_out_file_tmp.exists():
        msg = f"temporary output file {lh5_out_file_tmp} already exists"
        raise RuntimeError(msg)

    vert_it = LH5Iterator(lh5_in_file, "vtx", buffer_len=buffer_len)
    opti_it = LH5Iterator(lh5_in_file, "stp/optical", buffer_len=buffer_len)

    detectors = [str(d) for d in detectors]
    for d in detectors:
        if not d.isnumeric():
            log.warning("Detector ID %s is not numeric.", d)

    vert_df = None
    vert_df_bounds = None
    vert_it_count = 0
    hits_expected = 0

    def _store_vert_df():
        nonlocal vert_df
        if vert_df is None:
            return

        # sanity check that we did process all hits.
        hits_sum = 0
        for d in detectors:
            hits_sum += np.sum(vert_df[d])
        assert hits_sum == hits_expected

        log.info("store evt file %s (%d)", lh5_out_file_tmp, vert_it_count - 1)
        lh5.write(Table(vert_df), name=EVT_TABLE_NAME, lh5_file=lh5_out_file_tmp, wo_mode="append")
        vert_df = None

    # helper function for "windowed join". while iterating the optical hits, we have to
    # make sure that we always have the correct combined vertex/hit output table available.
    #
    # This function follows the assumption, that the output event ids are at least "somewhat"
    # monotonic, i.e. later chunks do not contain lower evtids than the previous chunk(s).
    # Going back is not implemented.
    def _ensure_vert_df(vert_it: LH5Iterator, evtid: int) -> None:
        nonlocal vert_df, vert_df_bounds, vert_it_count, hits_expected

        # skipping multiple chunks is possible in sparsely populated simulations.
        while vert_df_bounds is None or evtid > vert_df_bounds[1] or evtid < vert_df_bounds[0]:
            if vert_df_bounds is not None and vert_df is not None:
                if evtid < vert_df_bounds[0]:
                    msg = "non-monotonic evtid encountered, but cannot go back"
                    raise KeyError(msg)
                if evtid >= vert_df_bounds[0] and evtid <= vert_df_bounds[1]:
                    return  # vert_df already contains the given evtid.

            # here, evtid > vert_df_bounds[1] (or vert_df_bounds is still None). We need to fetch
            # the next event table chunk.

            vert_it_count += 1
            # we might have filled a dataframe, save it to disk.
            _store_vert_df()

            # read the next vertex chunk into memory.
            (vert_lgdo, vert_entry, vert_n_rows) = next(vert_it)
            vert_df = vert_lgdo.view_as("pd").iloc[0:vert_n_rows]

            # prepare vertex coordinates.
            vert_df = vert_df.set_index("evtid", drop=True).drop(["n_part", "time"], axis=1)
            vert_df_bounds = [vert_df.index.min(), vert_df.index.max()]
            hits_expected = 0
            # add columns for all detectors.
            for d in detectors:
                vert_df[d] = hit_count_type(0)

    log.info("prepare evt table")
    # use smaller integer type uint8 to spare RAM when storing types.
    hit_count_type = np.uint8
    for opti_it_count, (opti_lgdo, opti_entry, opti_n_rows) in enumerate(opti_it):
        assert (opti_it_count == 0) == (opti_entry == 0)
        opti_df = opti_lgdo.view_as("pd").iloc[0:opti_n_rows]

        log.info("build evt table (%d)", opti_it_count)
        for t in opti_df[["evtid", "det_uid"]].itertuples(name=None, index=False):
            _ensure_vert_df(vert_it, t[0])
            vert_df.loc[t[0], str(t[1])] += 1
            hits_expected += 1

    _store_vert_df()  # store the last chunk.

    # after finishing the output file, rename to the actual output file name.
    if lh5_out_file.exists():
        msg = f"output file {lh5_out_file_tmp} already exists after writing tmp output file"
        raise RuntimeError(msg)
    lh5_out_file_tmp.rename(lh5_out_file)


def read_optmap_evt(lh5_file: str, buffer_len: int = int(5e6)) -> LH5Iterator:
    return LH5Iterator(lh5_file, EVT_TABLE_NAME, buffer_len=buffer_len)



from __future__ import annotations

import logging
from pathlib import Path
from typing import Literal

import matplotlib.pyplot as plt
import numpy as np
from lgdo import lh5
from matplotlib import colors, widgets
from numpy.typing import NDArray

from .create import list_optical_maps

log = logging.getLogger(__name__)


def _get_weights(viewdata: dict):
    rolled = np.array([viewdata["axis"]] + [i for i in [2, 1, 0] if i != viewdata["axis"]])
    weights = viewdata["weights"].transpose(*rolled)

    extent_axes = [i for i in np.flip(rolled) if i != viewdata["axis"]]
    extent = [viewdata["edges"][i] for i in extent_axes]
    extent_d = [(e[1] - e[0]) / 2 for e in extent]
    extent = [(e[0] - extent_d[i], e[-1] + extent_d[i]) for i, e in enumerate(extent)]

    labels = [["x", "y", "z"][i] + " [m]" for i in extent_axes]

    return weights, np.array(extent).flatten(), labels


def _slice_text(viewdata: dict) -> str:
    axis_name = ["x", "y", "z"][viewdata["axis"]]
    axis_pos = viewdata["edges"][viewdata["axis"]][viewdata["idx"]]
    return f"{viewdata['detid']} | slice {axis_name} = {axis_pos:.2f} m"


def _process_key(event) -> None:
    fig = event.canvas.figure
    viewdata = fig.__reboost

    if event.key == "up":
        viewdata["axis"] = min(viewdata["axis"] + 1, 2)
    elif event.key == "down":
        viewdata["axis"] = max(viewdata["axis"] - 1, 0)
    elif event.key == "right":
        viewdata["idx"] += 1
    elif event.key == "left":
        viewdata["idx"] -= 1
    elif event.key == "c":
        _channel_selector(fig)

    max_idx = viewdata["weights"].shape[viewdata["axis"]] - 1
    viewdata["idx"] = max(min(viewdata["idx"], max_idx), 0)

    _update_figure(fig)


def _update_figure(fig) -> None:
    viewdata = fig.__reboost
    w, extent, labels = _get_weights(viewdata)

    ax = fig.axes[0]
    ax.texts[0].set_text(_slice_text(viewdata))
    ax.images[0].set_array(w[viewdata["idx"]])
    ax.images[0].set_extent(extent)
    ax.set_xlabel(labels[0])
    ax.set_ylabel(labels[1])
    ax.set_anchor("C")
    fig.canvas.draw()


def _channel_selector(fig) -> None:
    axbox = fig.add_axes([0.01, 0.01, 0.98, 0.98])
    channels = fig.__reboost["available_dets"]
    tb = widgets.RadioButtons(axbox, channels, active=channels.index(fig.__reboost["detid"]))

    def change_detector(label: str | None) -> None:
        if fig.__reboost["detid"] != label:
            fig.__reboost["detid"] = label
            edges, weights, _, _ = _prepare_data(*fig.__reboost["prepare_args"], label)
            fig.__reboost["weights"] = weights
            fig.__reboost["edges"] = edges
        tb.disconnect_events()
        axbox.remove()
        _update_figure(fig)

    tb.on_clicked(change_detector)
    fig.canvas.draw()


def _read_data(
    optmap_fn: str,
    detid: str = "all",
    histogram_choice: str = "p_det",
) -> tuple[tuple[NDArray], NDArray]:
    histogram = histogram_choice if histogram_choice != "p_det_err_rel" else "p_det"
    optmap_all = lh5.read(f"/{detid}/{histogram}", optmap_fn)
    optmap_edges = tuple([b.edges for b in optmap_all.binning])
    optmap_weights = optmap_all.weights.nda.copy()
    if histogram_choice == "p_det_err_rel":
        optmap_err = lh5.read(f"/{detid}/p_det_err", optmap_fn)
        divmask = optmap_weights > 0
        optmap_weights[divmask] = optmap_err.weights.nda[divmask] / optmap_weights[divmask]
        optmap_weights[~divmask] = -1

    return optmap_edges, optmap_weights


def _prepare_data(
    optmap_fn: str,
    divide_fn: str | None = None,
    cmap_min: float | Literal["auto"] = 1e-4,
    cmap_max: float | Literal["auto"] = 1e-2,
    histogram_choice: str = "p_det",
    detid: str = "all",
) -> tuple[tuple[NDArray], NDArray]:
    optmap_edges, optmap_weights = _read_data(optmap_fn, detid, histogram_choice)

    if divide_fn is not None:
        divide_edges, divide_map = _read_data(divide_fn, detid, histogram_choice)
        divmask = divide_map > 0
        optmap_weights[divmask] = optmap_weights[divmask] / divide_map[divmask]
        optmap_weights[~divmask] = -1

    if cmap_min == "auto":
        cmap_min = max(1e-10, optmap_weights[optmap_weights > 0].min())
    if cmap_max == "auto":
        cmap_max = optmap_weights.max()

    lower_count = np.sum((optmap_weights > 0) & (optmap_weights < cmap_min))
    if lower_count > 0:
        log.warning(
            "%d cells are non-zero and lower than the current colorbar minimum %.2e",
            lower_count,
            cmap_min,
        )
    higher_count = np.sum(optmap_weights > cmap_max)
    if higher_count > 0:
        log.warning(
            "%d cells are non-zero and higher than the current colorbar maximum %.2e",
            higher_count,
            cmap_max,
        )

    # set zero/close-to-zero values to a very small, but nonzero value. This means we can
    # style those cells using the `under` style and, can style `bad` (i.e. everything < 0
    # after this re-assignment) in a different way.
    optmap_weights[(optmap_weights >= 0) & (optmap_weights < 1e-50)] = min(1e-10, cmap_min)

    return optmap_edges, optmap_weights, cmap_min, cmap_max


def view_optmap(
    optmap_fn: list[str],
    detid: str = "all",
    divide_fn: str | None = None,
    start_axis: int = 2,
    cmap_min: float | Literal["auto"] = 1e-4,
    cmap_max: float | Literal["auto"] = 1e-2,
    histogram_choice: str = "p_det",
    title: str | None = None,
) -> None:
    available_dets = list_optical_maps(optmap_fn)

    prepare_args = (optmap_fn, divide_fn, cmap_min, cmap_max, histogram_choice)
    edges, weights, cmap_min, cmap_max = _prepare_data(*prepare_args, detid)

    fig = plt.figure(figsize=(10, 10))
    fig.canvas.mpl_connect("key_press_event", _process_key)
    start_axis_len = edges[start_axis].shape[0] - 1
    fig.__reboost = {
        "axis": start_axis,
        "weights": weights,
        "detid": detid,
        "edges": edges,
        "idx": min(int(start_axis_len / 2), start_axis_len - 1),
        "available_dets": available_dets,
        "prepare_args": prepare_args,
    }

    cmap = plt.cm.plasma.with_extremes(bad="w", under="gray", over="red")
    weights, extent, labels = _get_weights(fig.__reboost)
    plt.imshow(
        weights[fig.__reboost["idx"]],
        norm=colors.LogNorm(vmin=cmap_min, vmax=cmap_max),
        aspect=1,
        interpolation="none",
        cmap=cmap,
        extent=extent,
        origin="lower",
    )

    if title is None:
        title = Path(optmap_fn).stem
        if divide_fn is not None:
            title += " / " + Path(divide_fn).stem

    plt.suptitle(title)

    plt.xlabel(labels[0])
    plt.ylabel(labels[1])

    plt.text(0, 1.02, _slice_text(fig.__reboost), transform=fig.axes[0].transAxes)
    plt.colorbar()
    plt.show()



"""A program for combining the hits from various detectors, to build events.

Is able to parse a config file with the following format config file:

.. code-block:: yaml

    channels:
        geds_on:
        - det001
        - det002
        geds_ac:
        - det003

    outputs:
    - energy
    - multiplicity

    operations:
     energy_id:
        channels: geds_on
        aggregation_mode: gather
        query: "hit.energy > 25"
        expression: tcm.channel_id

     energy:
        aggregation_mode: keep_at_ch:evt.energy_id
        expression: "hit.energy > 25"
        channels: geds_on

     multiplicity:
        channels: geds_on
        aggregation_mode: sum
        expression: "hit.energy > 25"
        initial: 0


Must contain:
- "channels": dictionary of channel groupings
- "outputs": fields for the output file
- "operations": operations to perform see :func:`pygama.evt.build_evt.evaluate_expression` for more details.
"""

from __future__ import annotations

import logging

import awkward as ak
import numpy as np
from lgdo import Table
from lgdo.lh5 import LH5Iterator, write
from pygama.evt.build_evt import evaluate_expression
from pygama.evt.utils import TCMData

from . import utils

log = logging.getLogger(__name__)


def build_evt(
    hit_file: str, tcm_file: str, evt_file: str | None, config: dict, buffer: int = int(5e6)
) -> ak.Array | None:
    """Generates the event tier from the hit and tcm.

    Parameters
    ----------
    hit_file
        path to the hit tier file
    tcm_file
        path to the tcm tier file
    evt_file
        path to the evt tier (output) file, if `None` the :class:`Table` is returned in memory
    config
        dictionary of the configuration.
    buffer
        number of events to process simultaneously

    Returns
    -------
    ak.Array of the evt tier data (if the data is not saved to disk)
    """
    msg = "... beginning the evt tier processing"
    log.info(msg)

    # create the objects needed for evaluate expression

    file_info = {
        "hit": (hit_file, "hit", "det{:03}"),
        "evt": (evt_file, "evt"),
    }

    # iterate through the TCM

    out_ak = ak.Array([])
    mode = "of"

    # get channel groupings
    channels = {}
    for group, info in config["channels"].items():
        if isinstance(info, str):
            channels[group] = [info]

        elif isinstance(info, list):
            channels[group] = info

    for tcm_lh5, _, n_rows_read in LH5Iterator(tcm_file, "tcm", buffer_len=buffer):
        tcm_lh5_sel = tcm_lh5
        tcm_ak = tcm_lh5_sel.view_as("ak")[:n_rows_read]

        tcm = TCMData(
            id=np.array(ak.flatten(tcm_ak.array_id)),
            idx=np.array(ak.flatten(tcm_ak.array_idx)),
            cumulative_length=np.array(np.cumsum(ak.num(tcm_ak.array_id, axis=-1))),
        )

        n_rows = len(tcm.cumulative_length)
        out_tab = Table(size=n_rows)

        for name, info in config["operations"].items():
            msg = f"computing field {name}"
            log.debug(msg)

            defaultv = info.get("initial", np.nan)
            if isinstance(defaultv, str) and (defaultv in ["np.nan", "np.inf", "-np.inf"]):
                defaultv = eval(defaultv)

            channels_use = utils.get_channels_from_groups(info.get("channels", []), channels)
            channels_exclude = utils.get_channels_from_groups(
                info.get("exclude_channels", []), channels
            )

            if "aggregation_mode" not in info:
                field = out_tab.eval(
                    info["expression"].replace("evt.", ""), info.get("parameters", {})
                )
            else:
                field = evaluate_expression(
                    file_info,
                    tcm,
                    channels_use,
                    table=out_tab,
                    mode=info["aggregation_mode"],
                    expr=info["expression"],
                    query=info.get("query", None),
                    sorter=info.get("sort", None),
                    channels_skip=channels_exclude,
                    default_value=defaultv,
                    n_rows=n_rows,
                )

            msg = f"field {field}"
            log.debug(msg)
            out_tab.add_field(name, field)

        # remove fields if necessary
        existing_cols = list(out_tab.keys())
        for col in existing_cols:
            if col not in config["outputs"]:
                out_tab.remove_column(col, delete=True)

        # write
        if evt_file is not None:
            write(out_tab, "evt", evt_file, wo_mode=mode)
            mode = "append"
        else:
            out_ak = ak.concatenate((out_ak, out_tab.view_as("ak")))

    if evt_file is None:
        return out_ak
    return None



from __future__ import annotations

import time

from reboost.profile import ProfileDict


def test_profile():
    # create the object
    time_dict = ProfileDict()

    time_start = time.time()

    # create some fields
    time_dict.update_field("global_objects", time_start - 5)
    time_dict.update_field("processing_groups/germanium/read", time_start - 3)
    time_dict.update_field("processing_groups/germanium/detector_objects", time_start - 17)
    time_dict.update_field("processing_groups/germanium/write", time_start - 0.01)
    time_dict.update_field("processing_groups/germanium/operations/energy", time_start - 3)
    time_dict.update_field("processing_groups/germanium/operations/activeness", time_start - 42)
    time_dict.update_field("processing_groups/sipms/write", time_start - 12)

    # now update some fields
    time_dict.update_field("processing_groups/sipms/write", time_start - 7)

    assert round(time_dict["global_objects"], 1) == 5.0
    assert round(time_dict["processing_groups"]["germanium"]["read"], 1) == 3.0
    assert round(time_dict["processing_groups"]["germanium"]["detector_objects"], 1) == 17.0
    assert round(time_dict["processing_groups"]["germanium"]["write"], 1) == 0.0
    assert round(time_dict["processing_groups"]["germanium"]["operations"]["energy"], 1) == 3.0
    assert round(time_dict["processing_groups"]["germanium"]["operations"]["activeness"], 1) == 42.0
    assert round(time_dict["processing_groups"]["sipms"]["write"], 1) == 19.0



from __future__ import annotations

import awkward as ak
import numpy as np
import pytest
from lgdo import Array, VectorOfVectors

from reboost.math import functions, stats


def test_hpge_activeness():
    # test with VectorOfVectors and ak.Array input
    distances = ak.Array([[0.2], [0.6], [2]])

    activeness = functions.piecewise_linear_activeness(VectorOfVectors(distances), fccd=1, dlf=0.4)

    # first point should be 0
    assert activeness[0][0] == 0
    # second should be 0.1/0.5 = 0.2
    assert activeness[1][0] == pytest.approx(1 / 3.0)
    assert activeness[2][0] == 1

    # test with ak.Array input
    distances = ak.Array([[0.2], [0.6], [2]])
    activeness = functions.piecewise_linear_activeness(distances, fccd=1, dlf=0.4)

    # first point should be 0
    assert activeness[0][0] == 0
    # second should be 0.1/0.5 = 0.2
    assert activeness[1][0] == pytest.approx(1 / 3.0)
    assert activeness[2][0] == 1

    # test with Array
    activeness = functions.piecewise_linear_activeness([[0.2, 0.6, 2]], fccd=1, dlf=0.4)

    assert np.allclose(activeness.view_as("np"), [0, 1 / 3.0, 1])

    # test with array
    distances = ak.Array([0.2, 0.6, 2])
    activeness = functions.piecewise_linear_activeness(Array(distances), fccd=1, dlf=0.4)

    # first point should be 0
    assert activeness[0] == 0
    # second should be 0.1/0.5 = 0.2
    assert activeness[1] == pytest.approx(1 / 3.0)
    assert activeness[2] == 1


def test_vectorised_activeness():
    # vary fccd
    distances = ak.Array([[0.2, 5], [0.6], [2], [5]])
    edep = ak.Array([[100, 200], [100], [100], [100]])

    # simple case
    energy = functions.vectorised_active_energy(
        VectorOfVectors(distances), VectorOfVectors(edep), fccd=1, dlf=0.4
    )

    assert energy[0] == 200
    assert energy[1] == pytest.approx(100 / 3.0)
    assert energy[2] == 100

    # now vectorised over fccd
    energy_fccd = functions.vectorised_active_energy(
        VectorOfVectors(distances), VectorOfVectors(edep), fccd=[0, 0.5, 1, 2, 3], dlf=1
    )

    # with distance of 0.2 only fccd of 0 gives energy
    assert np.allclose(energy_fccd[0], [300, 200, 200, 200, 200])
    assert np.allclose(energy_fccd[1], [100, 100, 0, 0, 0])
    assert np.allclose(energy_fccd[2], [100, 100, 100, 0, 0])
    assert np.allclose(energy_fccd[3], [100, 100, 100, 100, 100])

    # vectorise over the dlf
    energy_dlf = functions.vectorised_active_energy(
        VectorOfVectors(distances), VectorOfVectors(edep), fccd=1, dlf=[0.4, 1]
    )

    assert np.allclose(energy_dlf[0], [200, 200])
    assert np.allclose(energy_dlf[1], [100 / 3.0, 0])
    assert np.allclose(energy_dlf[2], [100, 100])
    assert np.allclose(energy_dlf[3], [100, 100])


def test_sample():
    # list inputs
    samples = stats.gaussian_sample([1, 2, 3], [0.1, 0.1, 0.1])
    assert isinstance(samples, Array)

    # LGDO inputs
    samples = stats.gaussian_sample(Array(np.array([1, 2, 3])), Array(np.array([0.1, 0.1, 0.1])))
    assert isinstance(samples, Array)

    # ak inputs
    samples = stats.gaussian_sample(ak.Array([1, 2, 3]), ak.Array([1, 2, 3]))
    assert isinstance(samples, Array)

    # sigma float
    samples = stats.gaussian_sample([1, 2, 3], 0.1)
    assert isinstance(samples, Array)



from __future__ import annotations

from typing import Callable, NamedTuple

import lgdo
import numpy as np
import pint
from dbetto import AttrsDict
from lgdo import lh5
from scipy.interpolate import RegularGridInterpolator


class HPGeScalarRZField(NamedTuple):
    """A scalar field defined in the cylindrical-like (r, z) HPGe plane."""

    φ: Callable
    "Scalar field, function of the coordinates (r, z)."
    r_units: pint.Unit
    "Physical units of the coordinate `r`."
    z_units: pint.Unit
    "Physical units of the coordinate `z`."
    φ_units: pint.Unit
    "Physical units of the field."


def get_hpge_scalar_rz_field(
    filename: str, obj: str, field: str, out_of_bounds_val: int | float = np.nan, **kwargs
) -> HPGeScalarRZField:
    """Create an interpolator for a gridded scalar HPGe field defined on `(r, z)`.

    Reads from disk the following data structure: ::

        FILENAME/
        └── OBJ · struct{r,z,FIELD}
            ├── r · array<1>{real} ── {'units': 'UNITS'}
            ├── z · array<1>{real} ── {'units': 'UNITS'}
            └── FIELD · array<2>{real} ── {'units': 'UNITS'}

    where ``FILENAME``, ``OBJ`` and ``FIELD`` are provided as
    arguments to this function. `obj` is a :class:`~lgdo.types.struct.Struct`,
    `r` and `z` are one dimensional arrays specifying the radial and z
    coordinates of the rectangular grid — not the coordinates of each single
    grid point. In this coordinate system, the center of the p+ contact surface
    is at `(0, 0)`, with the p+ contact facing downwards. `field` is instead a
    two-dimensional array specifying the field value at each grid point. The
    first and second dimensions are `r` and `z`, respectively. NaN values are
    interpreted as points outside the detector profile in the `(r, z)` plane.

    Before returning a :class:`HPGeScalarRZField`, the gridded field is fed to
    :class:`scipy.interpolate.RegularGridInterpolator`.

    Parameters
    ----------
    filename
        name of the LH5 file containing the gridded scalar field.
    obj
        name of the HDF5 dataset where the data is saved.
    field
        name of the HDF5 dataset holding the field values.
    out_of_bounds_val
        value to use to replace NaNs in the field values.
    """
    data = lh5.read(obj, filename)

    if not isinstance(data, lgdo.Struct):
        msg = f"{obj} in {filename} is not an LGDO Struct"
        raise ValueError(msg)

    data = AttrsDict(
        {
            k: np.nan_to_num(data[k].view_as("np", with_units=True), nan=out_of_bounds_val)
            for k in ("r", "z", field)
        }
    )

    interpolator = RegularGridInterpolator((data.r.m, data.z.m), data[field].m, **kwargs)

    return HPGeScalarRZField(interpolator, data.r.u, data.z.u, data[field].u)



from __future__ import annotations

import logging

import awkward as ak
import legendhpges
import numpy as np
from lgdo import VectorOfVectors
from lgdo.types import LGDO
from numpy.typing import ArrayLike

log = logging.getLogger(__name__)


def distance_to_surface(
    positions_x: VectorOfVectors,
    positions_y: VectorOfVectors,
    positions_z: VectorOfVectors,
    hpge: legendhpges.base.HPGe,
    det_pos: ArrayLike,
    *,
    surface_type: str | None = None,
    unit: str = "mm",
    distances_precompute: VectorOfVectors | None = None,
    precompute_cutoff: float | None = None,
) -> VectorOfVectors:
    """Computes the distance from each step to the detector surface.

    The calculation can be performed for any surface type `nplus`, `pplus`,
    `passive` or `None`. In order to speed up the calculation we provide
    an option to only compute the distance for points within a certain distance
    of any surface (as computed by remage and stored in the "distances_precompute") argument.

    Parameters
    ----------
    positions_x
        Global x positions for each step.
    positions_y
        Global y positions for each step.
    positions_z
        Global z positions for each step.
    hpge
        HPGe object.
    det_pos
        position of the detector origin, must be a 3 component array corresponding to `(x,y,z)`.
    surface_type
        string of which surface to use, can be `nplus`, `pplus` `passive` or None (in which case the distance to any surface is calculated).
    unit
        unit for the hit tier positions table.
    distances_precompute
        VectorOfVectors of distance to any surface computed by remage.
    precompute_cutoff
        cutoff on distances_precompute to not compute the distance for (in mm)

    Returns
    -------
    VectorOfVectors with the same shape as `positions_x/y/z` of the distance to the surface.

    Note
    ----
    `positions_x/positions_y/positions_z` must all have the same shape.
    """
    factor = np.array([1, 100, 1000])[unit == np.array(["mm", "cm", "m"])][0]

    # compute local positions
    pos = []
    sizes = []

    for idx, pos_tmp in enumerate([positions_x, positions_y, positions_z]):
        local_pos_tmp = ak.Array(pos_tmp) * factor - det_pos[idx]
        local_pos_flat_tmp = ak.flatten(local_pos_tmp).to_numpy()
        pos.append(local_pos_flat_tmp)
        sizes.append(ak.num(local_pos_tmp, axis=1))

    if not ak.all(sizes[0] == sizes[1]) or not ak.all(sizes[0] == sizes[2]):
        msg = "all position vector of vector must have the same shape"
        raise ValueError(msg)

    size = sizes[0]
    # restructure the positions
    local_positions = np.vstack(pos).T

    # get indices
    surface_indices = (
        np.where(np.array(hpge.surfaces) == surface_type) if surface_type is not None else None
    )

    # distance calc itself
    if distances_precompute is None:
        distances = hpge.distance_to_surface(local_positions, surface_indices=surface_indices)
    else:
        # decide when the calculation needs to be run
        if isinstance(distances_precompute, LGDO):
            distances_precompute = distances_precompute.view_as("ak")

        distances_precompute_flat = ak.flatten(distances_precompute)
        distances = np.full_like(distances_precompute_flat.to_numpy(), np.nan, dtype=float)

        # values to compute
        indices = distances_precompute_flat < precompute_cutoff

        # compute the distances
        distances[indices] = hpge.distance_to_surface(
            local_positions[indices], surface_indices=surface_indices
        )

    return VectorOfVectors(ak.unflatten(distances, size))



from __future__ import annotations

import json
from pathlib import Path

import awkward as ak
import legendhpges
import numpy as np
import pyg4ometry as pg4
import pygeomtools
import pytest
from dbetto import AttrsDict
from legendtestdata import LegendTestData
from lgdo import Array, Table

import reboost


@pytest.fixture(scope="session")
def test_data_configs():
    ldata = LegendTestData()
    ldata.checkout("5f9b368")
    return ldata.get_path("legend/metadata/hardware/detectors/germanium/diodes")


@pytest.fixture(scope="session")
def make_gdml(test_data_configs):
    reg = pg4.geant4.Registry()
    meta = f"{test_data_configs}/C99000A.json"
    hpge = legendhpges.make_hpge(meta, registry=reg)

    world_s = pg4.geant4.solid.Orb("World_s", 20, registry=reg, lunit="cm")
    world_l = pg4.geant4.LogicalVolume(world_s, "G4_Galactic", "World", registry=reg)
    reg.setWorld(world_l)

    # now place the two HPGe detectors in the argon
    hpge_p = pg4.geant4.PhysicalVolume(
        [0, 0, 0], [5, 0, 0, "cm"], hpge, "HPGE", world_l, registry=reg
    )
    with Path(meta).open() as file:
        metadata = json.load(file)

    hpge_p.pygeom_active_detector = pygeomtools.RemageDetectorInfo(
        "germanium",
        1,
        metadata,
    )
    pygeomtools.detectors.write_detector_auxvals(reg)

    w = pg4.gdml.Writer()
    w.addDetector(reg)
    w.write(f"{test_data_configs}/geometry.gdml")

    return f"{test_data_configs}/geometry.gdml"


def test_get_objects(test_data_configs, make_gdml):
    # check basic eval
    expression = "pyg4ometry.geant4.Registry()"
    reg = reboost.core.evaluate_object(expression, {})
    assert reg is not None
    objects = AttrsDict({"reg": reg})

    # get legend-test data
    # check if we can make a HPGe
    path = test_data_configs + r"/{DETECTOR}"
    expression = f"legendhpges.make_hpge(f'{path}.json',registry = OBJECTS.reg)"
    assert reboost.core.evaluate_object(expression, {"DETECTOR": "C99000A", "OBJECTS": objects})

    # test construction of the global objects
    # first not involving any args, second with ARGS
    args = AttrsDict(
        {
            "metadata": test_data_configs,
            "gdml": make_gdml,
            "name": "test",
        }
    )

    obj = reboost.core.get_global_objects(
        {
            "geometry": "pyg4ometry.gdml.Reader(ARGS.gdml).getRegistry()",
            "reg": "pyg4ometry.geant4.Registry()",
            "test_data": "legendtestdata.LegendTestData()",
            "meta": "dbetto.TextDB(ARGS.metadata)",
        },
        local_dict={"ARGS": args},
    )
    # check it worked
    assert sorted(obj.keys()) == ["geometry", "meta", "reg", "test_data"]

    # test getting detector objects
    expressions = {
        "meta": "pygeomtools.get_sensvol_metadata(OBJECTS.geometry, DETECTOR)",
        "hpge": "legendhpges.make_hpge(pygeomtools.get_sensvol"
        "_metadata(OBJECTS.geometry, DETECTOR),registry = OBJECTS.reg,name=ARGS.name)",
    }

    det_obj = reboost.core.get_detector_objects(["HPGE"], expressions, args, obj)

    assert next(iter(det_obj.keys())) == "HPGE"
    assert list(det_obj["HPGE"].keys()) == ["meta", "hpge"]


def test_eval_expression():
    hit_table = Table()
    hit_table.add_field("field", Array([1, 2, 3]))
    hit_table.add_field("other", Array([2, 4, 5]))
    hit_table.add_field("distances", Array([0.1, 0.6, 3]))

    # evaluate basic
    expression = "HITS.field + HITS.other"
    assert np.all(
        reboost.core.evaluate_output_column(hit_table, expression, {}, table_name="HITS").view_as(
            "np"
        )
        == [3, 6, 8]
    )

    # test also with locals
    expression = "HITS.field*args.val"
    local = {"args": AttrsDict({"val": 2})}
    assert np.all(
        reboost.core.evaluate_output_column(
            hit_table, expression, local, table_name="HITS"
        ).view_as("np")
        == [2, 4, 6]
    )

    # test with a reboost function
    expression = "reboost.math.functions.piecewise_linear_activeness(HITS.distances, fccd=args.fccd, dlf=args.dlf)"

    local = {"args": AttrsDict({"fccd": 1, "dlf": 0.5})}

    assert np.allclose(
        reboost.core.evaluate_output_column(
            hit_table, expression, local, table_name="HITS"
        ).view_as("np"),
        [0, 0.2, 1],
    )


def test_detector_mapping():
    # basic
    assert reboost.core.get_detectors_mapping("[str(i) for i in range(3)]") == {
        "0": ["0"],
        "1": ["1"],
        "2": ["2"],
    }

    assert reboost.core.get_detectors_mapping("0") == {"0": ["0"]}

    # with input name
    assert reboost.core.get_detectors_mapping(
        "[str(i) for i in range(3)]", input_detector_name="dets"
    ) == {"dets": ["0", "1", "2"]}

    # with  objects:
    objs = AttrsDict({"format": "ch"})
    assert reboost.core.get_detectors_mapping(
        "[f'{OBJECTS.format}{i}' for i in range(3)]", input_detector_name="dets", objects=objs
    ) == {"dets": ["ch0", "ch1", "ch2"]}


def test_remove_columns():
    tab = Table({"a": Array([1, 2, 3]), "b": Array([1, 2, 3])})

    tab = reboost.core.remove_columns(tab, ["a"])

    assert next(iter(tab.keys())) == "a"


def test_merge():
    tab = Table({"a": Array([2, 3])})
    output_tab = ak.Array({"a": [0, 1]})

    orig = reboost.core.merge(tab, None)
    assert np.all(orig.a == [2, 3])

    merged = reboost.core.merge(tab, output_tab)

    assert np.all(merged.a == [0, 1, 2, 3])
